from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from great_expectations.expectations.expectation import (
    BatchExpectation,
    render_suite_parameter_string,
)
from great_expectations.render import (
    LegacyDiagnosticRendererType,
    LegacyRendererType,
    RenderedStringTemplateContent,
    RenderedTableContent,
)
from great_expectations.render.renderer.renderer import renderer
from great_expectations.render.renderer_configuration import (
    RendererConfiguration,
    RendererValueType,
)
from great_expectations.render.util import num_to_str, substitute_none_for_missing

if TYPE_CHECKING:
    from great_expectations.core import (
        ExpectationValidationResult,
    )
    from great_expectations.expectations.expectation_configuration import (
        ExpectationConfiguration,
    )
    from great_expectations.render.renderer_configuration import AddParamArgs


# NOTE: This Expectation is incomplete and not ready for use.
#       It should remain unexported until it meets the requirements set by our V1 API.
class ExpectColumnPairCramersPhiValueToBeLessThan(BatchExpectation):
    def __init__(self, *args, **kwargs):
        raise NotImplementedError

    library_metadata = {
        "maturity": "production",
        "tags": [
            "core expectation",
            "multi-column expectation",
            "needs migration to modular expectations api",
        ],
        "contributors": ["@great_expectations"],
        "requirements": [],
    }

    metric_dependencies = tuple()
    success_keys = (
        "column_A",
        "column_B",
        "threshold",
    )
    # default_kwarg_values = {
    #     "column_A": None,
    #     "column_B": None,
    #     "bins_A": None,
    #     "bins_B": None,
    #     "n_bins_A": None,
    #     "n_bins_B": None,
    #     "threshold": 0.1,
    #     "result_format": "BASIC",
    #     "catch_exceptions": False,
    # }
    args_keys = (
        "column_A",
        "column_B",
    )

    @classmethod
    def _prescriptive_template(
        cls,
        renderer_configuration: RendererConfiguration,
    ) -> RendererConfiguration:
        add_param_args: AddParamArgs = (
            ("column_A", RendererValueType.STRING),
            ("column_B", RendererValueType.STRING),
        )
        for name, param_type in add_param_args:
            renderer_configuration.add_param(name=name, param_type=param_type)

        params = renderer_configuration.params

        if not params.column_A or not params.column_B:
            renderer_configuration.template_str = " unrecognized kwargs for expect_column_pair_cramers_phi_value_to_be_less_than: missing column."  # noqa: E501 # FIXME CoP
        else:
            renderer_configuration.template_str = (
                "Values in $column_A and $column_B must be independent."
            )

        return renderer_configuration

    @classmethod
    @renderer(renderer_type=LegacyRendererType.PRESCRIPTIVE)
    @render_suite_parameter_string
    def _prescriptive_renderer(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
        **kwargs,
    ):
        runtime_configuration = runtime_configuration or {}
        _ = runtime_configuration.get("include_column_name") is not False
        styling = runtime_configuration.get("styling")
        params = substitute_none_for_missing(configuration.kwargs, ["column_A", "column_B"])
        if (params["column_A"] is None) or (params["column_B"] is None):
            template_str = " unrecognized kwargs for expect_column_pair_cramers_phi_value_to_be_less_than: missing column."  # noqa: E501 # FIXME CoP
        else:
            template_str = "Values in $column_A and $column_B must be independent."

        rendered_string_template_content = RenderedStringTemplateContent(
            **{
                "content_block_type": "string_template",
                "string_template": {
                    "template": template_str,
                    "params": params,
                    "styling": styling,
                },
            }
        )

        return [rendered_string_template_content]

    @classmethod
    @renderer(renderer_type=LegacyDiagnosticRendererType.OBSERVED_VALUE)
    def _diagnostic_observed_value_renderer(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
        **kwargs,
    ):
        observed_value = result.result.get("observed_value")
        column_A = result.expectation_config.kwargs["column_A"]
        column_B = result.expectation_config.kwargs["column_B"]
        crosstab = result.result.get("details", {}).get("crosstab")

        if observed_value is not None:
            observed_value = num_to_str(observed_value, precision=3, use_locale=True)
            if crosstab is not None:
                table = [[""] + list(crosstab.columns)]
                for col in range(len(crosstab)):
                    table.append([crosstab.index[col]] + list(crosstab.iloc[col, :]))

                return RenderedTableContent(
                    **{
                        "content_block_type": "table",
                        "header": f"Observed cramers phi of {observed_value}. \n"
                        f"Crosstab between {column_A} (rows) and {column_B} (columns):",
                        "table": table,
                        "styling": {
                            "body": {
                                "classes": [
                                    "table",
                                    "table-sm",
                                    "table-unbordered",
                                    "col-4",
                                    "mt-2",
                                ],
                            }
                        },
                    }
                )
            else:
                return observed_value
        else:
            return "--"



from __future__ import annotations

from typing import TYPE_CHECKING, Iterable

from great_expectations._docs_decorators import public_api
from great_expectations.analytics.client import submit as submit_event
from great_expectations.analytics.events import (
    ActionInfo,
    CheckpointCreatedEvent,
    CheckpointDeletedEvent,
)
from great_expectations.checkpoint.checkpoint import Checkpoint
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.factory.factory import Factory
from great_expectations.exceptions import DataContextError

if TYPE_CHECKING:
    from great_expectations import ValidationDefinition
    from great_expectations.core.data_context_key import StringKey
    from great_expectations.data_context.store.checkpoint_store import (
        CheckpointStore,
    )
    from great_expectations.data_context.types.resource_identifiers import GXCloudIdentifier


@public_api
class CheckpointFactory(Factory[Checkpoint]):
    """
    Responsible for basic CRUD operations on a Data Context's Checkpoints.
    """

    def __init__(self, store: CheckpointStore):
        self._store = store

    @public_api
    @override
    def add(self, checkpoint: Checkpoint) -> Checkpoint:
        """Add a Checkpoint to the collection.

        Parameters:
            checkpoint: Checkpoint to add

        Raises:
            DataContextError: if Checkpoint already exists
        """
        key = self._store.get_key(name=checkpoint.name, id=None)
        if self._store.has_key(key=key):
            raise DataContextError(  # noqa: TRY003 # FIXME CoP
                f"Cannot add Checkpoint with name {checkpoint.name} because it already exists."
            )

        self._store.add(key=key, value=checkpoint)

        # TODO: Add id adding logic to CheckpointStore to prevent round trip
        persisted_checkpoint = self._get(key=key)

        submit_event(
            event=CheckpointCreatedEvent(
                checkpoint_id=persisted_checkpoint.id,
                validation_definition_ids=[
                    validation_definition.id
                    for validation_definition in checkpoint.validation_definitions
                ],
                actions=[
                    ActionInfo(
                        type=action.type,
                        # notify_on is not a property of all Actions
                        notify_on=getattr(action, "notify_on", None),
                    )
                    for action in checkpoint.actions
                ],
            )
        )

        return persisted_checkpoint

    @public_api
    @override
    def delete(self, name: str) -> None:
        """Delete a Checkpoint from the collection.

        Parameters:
            name: The name of the Checkpoint to delete

        Raises:
            DataContextError: if Checkpoint doesn't exist
        """
        try:
            checkpoint = self.get(name=name)
        except DataContextError as e:
            raise DataContextError(  # noqa: TRY003 # FIXME CoP
                f"Cannot delete Checkpoint with name {name} because it cannot be found."
            ) from e

        key = self._store.get_key(name=checkpoint.name, id=checkpoint.id)
        self._store.remove_key(key=key)

        submit_event(
            event=CheckpointDeletedEvent(
                checkpoint_id=checkpoint.id,
            )
        )

    @public_api
    @override
    def get(self, name: str) -> Checkpoint:
        """Get a Checkpoint from the collection by name.

        Parameters:
            name: Name of Checkpoint to get

        Raises:
            DataContextError: when Checkpoint is not found.
        """
        key = self._store.get_key(name=name, id=None)
        if not self._store.has_key(key=key):
            raise DataContextError(f"Checkpoint with name {name} was not found.")  # noqa: TRY003 # FIXME CoP

        return self._get(key=key)

    @public_api
    @override
    def all(self) -> Iterable[Checkpoint]:
        """Get all Checkpoints."""
        return self._store.get_all()

    def _get(self, key: GXCloudIdentifier | StringKey) -> Checkpoint:
        checkpoint = self._store.get(key=key)
        if not isinstance(checkpoint, Checkpoint):
            raise ValueError(f"Object with key {key} was found, but it is not a Checkpoint.")  # noqa: TRY003, TRY004 # FIXME CoP

        return checkpoint

    @public_api
    @override
    def add_or_update(self, checkpoint: Checkpoint) -> Checkpoint:
        """Add or update a Checkpoint by name.

        If a Checkpoint with the same name exists, overwrite it, otherwise
        create a new Checkpoint.

        Args:
            checkpoint: Checkpoint to add or update
        """

        try:
            existing_checkpoint = self.get(name=checkpoint.name)
        except DataContextError:
            # checkpoint doesn't exist yet, so add it
            self._add_or_update_validation_definitions(
                validation_definitions=checkpoint.validation_definitions,
                existing_validation_definitions=[],
            )
            return self.add(checkpoint=checkpoint)

        # update checkpoint
        checkpoint.id = existing_checkpoint.id
        self._add_or_update_validation_definitions(
            validation_definitions=checkpoint.validation_definitions,
            existing_validation_definitions=existing_checkpoint.validation_definitions,
        )
        checkpoint.save()
        return checkpoint

    def _add_or_update_validation_definitions(
        self,
        validation_definitions: list[ValidationDefinition],
        existing_validation_definitions: list[ValidationDefinition],
    ):
        from great_expectations.data_context import project_manager

        val_def_ids_by_name = {
            val_def.name: val_def.id for val_def in existing_validation_definitions
        }
        val_def_factory = project_manager.get_validation_definitions_factory()
        for val_def in validation_definitions:
            if val_def.name in val_def_ids_by_name:
                val_def.id = val_def_ids_by_name[val_def.name]
            val_def_factory.add_or_update(validation=val_def)



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Any,
    Callable,
    Dict,
    Generic,
    Mapping,
    Optional,
    Union,
)

from great_expectations.compatibility import pydantic
from great_expectations.compatibility.pydantic import Field, StrictStr
from great_expectations.compatibility.pydantic import json as pydantic_json
from great_expectations.compatibility.pydantic import (
    schema as pydantic_schema,
)

# default_ref_template
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.batch_definition import PartitionerT

# moving this import into TYPE_CHECKING requires forward refs to be updated.
from great_expectations.datasource.fluent.data_connector.batch_filter import (
    BatchSlice,
    parse_batch_slice,
)

if TYPE_CHECKING:
    from typing_extensions import TypeAlias

    MappingIntStrAny = Mapping[Union[int, str], Any]
    AbstractSetIntStr = AbstractSet[Union[int, str]]


# BatchParameters is a dict that is composed into a BatchRequest that specifies the
# Batches one wants as returned. The keys represent dimensions one can filter the data along
# and the values are the realized. If a value is None or unspecified, the batch_request
# will capture all data along this dimension. For example, if we have a year and month
# partitioner, and we want to query all months in the year 2020, the batch parameters
# would look like:
#   options = { "year": 2020 }
BatchParameters: TypeAlias = Dict[StrictStr, Any]


class BatchRequest(pydantic.GenericModel, Generic[PartitionerT]):
    """A BatchRequest is the way to specify which data Great Expectations will validate.

    A Batch Request is provided to a Data Asset in order to create one or more Batches.

    Args:
        datasource_name: The name of the Datasource used to connect to the data.
        data_asset_name: The name of the Data Asset used to connect to the data.
        options: A dict that can be used to filter the batch groups associated with the Data Asset.
            The dict structure depends on the asset type. The available keys for dict can be obtained by
            calling DataAsset.get_batch_parameters_keys(...).
        batch_slice: A python slice that can be used to filter the sorted batches by index.
            e.g. `batch_slice = "[-5:]"` will request only the last 5 batches after the options filter is applied.

    Returns:
        BatchRequest
    """  # noqa: E501 # FIXME CoP

    datasource_name: StrictStr = Field(
        ...,
        allow_mutation=False,
        description="The name of the Datasource used to connect to the data.",
    )
    data_asset_name: StrictStr = Field(
        ...,
        allow_mutation=False,
        description="The name of the Data Asset used to connect to the data.",
    )
    options: BatchParameters = Field(
        default_factory=dict,
        allow_mutation=True,
        description=(
            "A map that can be used to filter the batch groups associated with the Data Asset. "
            "The structure and types depends on the asset type."
        ),
    )
    partitioner: Optional[PartitionerT] = None
    _batch_slice_input: Optional[BatchSlice] = pydantic.PrivateAttr(
        default=None,
    )

    def __init__(self, **kwargs) -> None:
        _batch_slice_input: Optional[BatchSlice] = None
        if "batch_slice" in kwargs:
            _batch_slice_input = kwargs.pop("batch_slice")
        super().__init__(**kwargs)
        self._batch_slice_input = _batch_slice_input

    @property
    def batch_slice(self) -> slice:
        """A built-in slice that can be used to filter a list of batches by index."""
        return parse_batch_slice(batch_slice=self._batch_slice_input)

    def update_batch_slice(self, value: Optional[BatchSlice] = None) -> None:
        """Updates the batch_slice on this BatchRequest.

        Args:
            value: The new value to be parsed into a python slice and set on the batch_slice attribute.

        Returns:
            None
        """  # noqa: E501 # FIXME CoP
        try:
            parse_batch_slice(batch_slice=value)
        except (TypeError, ValueError) as e:
            raise ValueError(f"Failed to parse BatchSlice to slice: {e}")  # noqa: TRY003 # FIXME CoP
        self._batch_slice_input = value

    class Config:
        extra = pydantic.Extra.forbid
        property_set_methods = {"batch_slice": "update_batch_slice"}
        validate_assignment = True

    @override
    def __setattr__(self, key, val):
        method = self.__config__.property_set_methods.get(key)
        if method is None:
            super().__setattr__(key, val)
        else:
            getattr(self, method)(val)

    @pydantic.validator("options", pre=True)
    def _validate_options(cls, options) -> BatchParameters:
        if options is None:
            return {}
        if not isinstance(options, dict):
            raise TypeError("BatchParameters must take the form of a dictionary.")  # noqa: TRY003 # FIXME CoP
        if any(not isinstance(key, str) for key in options):
            raise TypeError("BatchParameters keys must all be strings.")  # noqa: TRY003 # FIXME CoP
        return options

    @override
    def json(  # noqa: PLR0913 # FIXME CoP
        self,
        *,
        include: Optional[Union[AbstractSetIntStr, MappingIntStrAny]] = None,
        exclude: Optional[Union[AbstractSetIntStr, MappingIntStrAny]] = None,
        by_alias: bool = False,
        skip_defaults: Optional[bool] = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        encoder: Optional[Callable[[Any], Any]] = None,
        models_as_dict: bool = True,
        **dumps_kwargs: Any,
    ) -> str:
        """
        Generate a json representation of the BatchRequest, optionally specifying which
        fields to include or exclude.
        """
        return super().json(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            skip_defaults=skip_defaults,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            encoder=encoder,
            models_as_dict=models_as_dict,
            **dumps_kwargs,
        )

    @override
    def dict(  # noqa: PLR0913 # FIXME CoP
        self,
        *,
        include: AbstractSetIntStr | MappingIntStrAny | None = None,
        exclude: AbstractSetIntStr | MappingIntStrAny | None = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        # deprecated - use exclude_unset instead
        skip_defaults: bool | None = None,
    ) -> dict[str, Any]:
        """
        Generate a dictionary representation of the BatchRequest, optionally specifying which
        fields to include or exclude.
        """
        # batch_slice is only a property/pydantic setter, so we need to add a field
        # if we want it to show up in dict() with the _batch_request_input
        self.__fields__["batch_slice"] = pydantic.fields.ModelField(
            name="batch_slice",
            type_=Optional[BatchSlice],  # type: ignore[arg-type] # FIXME CoP
            required=False,
            default=None,
            model_config=self.__config__,
            class_validators=None,
        )
        property_set_methods = self.__config__.property_set_methods  # type: ignore[attr-defined] # FIXME CoP
        self.__config__.property_set_methods = {}  # type: ignore[attr-defined] # FIXME CoP
        self.__setattr__("batch_slice", self._batch_slice_input)
        result = super().dict(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            skip_defaults=skip_defaults,
        )
        # revert model changes
        self.__config__.property_set_methods = property_set_methods  # type: ignore[attr-defined] # FIXME CoP
        self.__fields__.pop("batch_slice")
        return result

    @classmethod
    @override
    def schema_json(
        cls,
        *,
        by_alias: bool = True,
        ref_template: str = pydantic_schema.default_ref_template,
        **dumps_kwargs: Any,
    ) -> str:
        # batch_slice is only a property/pydantic setter, so we need to add a field
        # if we want its definition to show up in schema_json()
        cls.__fields__["batch_slice"] = pydantic.fields.ModelField(
            name="batch_slice",
            type_=Optional[BatchSlice],  # type: ignore[arg-type] # FIXME CoP
            required=False,
            default=None,
            model_config=cls.__config__,
            class_validators=None,
        )
        result = cls.__config__.json_dumps(
            cls.schema(by_alias=by_alias, ref_template=ref_template),
            default=pydantic_json.pydantic_encoder,
            **dumps_kwargs,
        )
        # revert model changes
        cls.__fields__.pop("batch_slice")
        return result



import json
import uuid
from unittest import mock

import pact
import pytest
import requests

import great_expectations as gx
from great_expectations.data_context import CloudDataContext
from great_expectations.datasource.fluent import PandasDatasource
from tests.integration.cloud.rest_contracts.conftest import JsonData, PactBody
from tests.integration.cloud.rest_contracts.test_data_context_configuration import (
    GET_DATA_CONTEXT_CONFIGURATION_MIN_RESPONSE_BODY,
)
from tests.integration.cloud.rest_contracts.test_datasource import (
    GET_DATASOURCE_MIN_RESPONSE_BODY,
    POST_DATASOURCE_MIN_RESPONSE_BODY,
)


def _convert_matcher_to_value(matcher: pact.matchers.Matcher) -> JsonData:
    return matcher.generate()["contents"]


def _reify_pact_body(  # noqa: C901 # FIXME CoP
    body: PactBody,
) -> JsonData:
    if isinstance(body, list):
        for index, item in enumerate(body):
            if isinstance(item, pact.matchers.Matcher):
                body[index] = _convert_matcher_to_value(matcher=item)
            body[index] = _reify_pact_body(body=body[index])
        return body
    elif isinstance(body, pact.matchers.Matcher):
        return _reify_pact_body(body=_convert_matcher_to_value(matcher=body))
    elif isinstance(body, dict):
        # calling generate on a parent matcher causes all child matchers
        # to take the form of a dict with the following keys
        if all(key in body for key in ("json_class", "data")):
            body = body["data"]["generate"]
        else:
            for key, value in body.items():
                if isinstance(value, pact.matchers.Matcher):
                    body[key] = _convert_matcher_to_value(matcher=value)
                body[key] = _reify_pact_body(body=body[key])
        return body
    else:
        return body


def _get_mock_response_from_pact_response_body(
    status_code: int,
    pact_response_body: PactBody,
) -> requests.Response:
    response_body: JsonData = _reify_pact_body(
        body=pact_response_body,
    )
    mock_response = requests.Response()
    mock_response.status_code = status_code
    mock_response._content = json.dumps(response_body).encode("utf-8")
    return mock_response


@pytest.fixture
def mock_cloud_data_context() -> CloudDataContext:
    mock_response: requests.Response = _get_mock_response_from_pact_response_body(
        status_code=200,
        pact_response_body=GET_DATA_CONTEXT_CONFIGURATION_MIN_RESPONSE_BODY,
    )

    with mock.patch(
        target="requests.Session.get",
        return_value=mock_response,
    ):
        mock_cloud_data_context: CloudDataContext = gx.get_context(
            mode="cloud",
            cloud_base_url="https://fake-host.io",
            cloud_organization_id=str(uuid.uuid4()),
            cloud_access_token="not a real token",
        )

    assert isinstance(mock_cloud_data_context, CloudDataContext)
    return mock_cloud_data_context


@pytest.fixture
def mock_cloud_pandas_datasource(
    mock_cloud_data_context: CloudDataContext,
) -> PandasDatasource:
    datasource_name = "mock_cloud_pandas_datasource"

    mock_post_response: requests.Response = _get_mock_response_from_pact_response_body(
        status_code=200,
        pact_response_body=POST_DATASOURCE_MIN_RESPONSE_BODY,
    )

    mock_get_response: requests.Response = _get_mock_response_from_pact_response_body(
        status_code=200,
        pact_response_body=GET_DATASOURCE_MIN_RESPONSE_BODY,
    )

    with mock.patch(
        target="great_expectations.core.datasource_dict.DatasourceDict.data",
        return_value={},
    ):
        with mock.patch(
            target="requests.Session.post",
            return_value=mock_post_response,
        ):
            with mock.patch(
                target="requests.Session.get",
                return_value=mock_get_response,
            ):
                mock_cloud_pandas_datasource: PandasDatasource = (
                    mock_cloud_data_context.data_sources.add_pandas(name=datasource_name)
                )

    assert isinstance(mock_cloud_pandas_datasource, PandasDatasource)
    assert mock_cloud_pandas_datasource.name == datasource_name

    return mock_cloud_pandas_datasource



from great_expectations.metrics.column import ColumnMetric
from great_expectations.metrics.metric_results import MetricResult


class ColumnDistinctValuesCountResult(MetricResult[int]): ...


class ColumnDistinctValuesCount(ColumnMetric[ColumnDistinctValuesCountResult]):
    """Count of distinct values in a column"""

    name = "column.distinct_values.count"



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Optional

from great_expectations.compatibility.typing_extensions import override
from great_expectations.expectations.metrics.map_metric_provider.map_metric_provider import (
    MapMetricProvider,
)
from great_expectations.validator.metric_configuration import MetricConfiguration

if TYPE_CHECKING:
    from great_expectations.execution_engine import ExecutionEngine
    from great_expectations.expectations.expectation_configuration import (
        ExpectationConfiguration,
    )

logger = logging.getLogger(__name__)


class ColumnMapMetricProvider(MapMetricProvider):
    """Defines metrics that are evaluated for every row for a single column. An example of a column map
    metric is `column_values.null` (which is implemented as a `ColumnMapMetricProvider`, a subclass of
    `MapMetricProvider`).

    ---Documentation---
        - https://docs.greatexpectations.io/docs/guides/expectations/creating_custom_expectations/how_to_create_custom_column_map_expectations
    """  # noqa: E501 # FIXME CoP

    condition_domain_keys = (
        "batch_id",
        "table",
        "column",
        "row_condition",
        "condition_parser",
    )
    function_domain_keys = (
        "batch_id",
        "table",
        "column",
        "row_condition",
        "condition_parser",
    )
    condition_value_keys = tuple()
    function_value_keys = tuple()

    @classmethod
    @override
    def _get_evaluation_dependencies(
        cls,
        metric: MetricConfiguration,
        configuration: Optional[ExpectationConfiguration] = None,
        execution_engine: Optional[ExecutionEngine] = None,
        runtime_configuration: Optional[dict] = None,
    ):
        dependencies: dict = super()._get_evaluation_dependencies(
            metric=metric,
            configuration=configuration,
            execution_engine=execution_engine,
            runtime_configuration=runtime_configuration,
        )
        table_domain_kwargs: dict = {
            k: v for k, v in metric.metric_domain_kwargs.items() if k != "column"
        }
        dependencies["table.column_types"] = MetricConfiguration(
            metric_name="table.column_types",
            metric_domain_kwargs=table_domain_kwargs,
            metric_value_kwargs={
                "include_nested": True,
            },
        )
        dependencies["table.columns"] = MetricConfiguration(
            metric_name="table.columns",
            metric_domain_kwargs=table_domain_kwargs,
            metric_value_kwargs=None,
        )
        dependencies["table.row_count"] = MetricConfiguration(
            metric_name="table.row_count",
            metric_domain_kwargs=table_domain_kwargs,
            metric_value_kwargs=None,
        )
        return dependencies



from __future__ import annotations

import logging
from functools import wraps
from typing import (
    Any,
    Callable,
    Dict,
    Type,
)

from great_expectations.compatibility.sqlalchemy import sqlalchemy as sa
from great_expectations.core.metric_domain_types import MetricDomainTypes
from great_expectations.core.metric_function_types import (
    MetricPartialFunctionTypes,
)
from great_expectations.execution_engine import (
    ExecutionEngine,
    PandasExecutionEngine,
    SparkDFExecutionEngine,
    SqlAlchemyExecutionEngine,
)
from great_expectations.expectations.metrics.metric_provider import (
    metric_partial,
)
from great_expectations.expectations.metrics.util import (
    get_dbms_compatible_metric_domain_kwargs,
)

logger = logging.getLogger(__name__)


def column_pair_function_partial(  # noqa: C901 #  16
    engine: Type[ExecutionEngine],
    partial_fn_type: MetricPartialFunctionTypes | None = None,
    **kwargs,
):
    """Provides engine-specific support for authoring a metric_fn with a simplified signature.

    A metric function that is decorated as a column_pair_function_partial will be called with the engine-specific
    column_list type and any value_kwargs associated with the Metric for which the provider function is being declared.

    Args:
        engine: The `ExecutionEngine` used to to evaluate the condition
        partial_fn_type: The metric function
        **kwargs: Arguments passed to specified function

    Returns:
        An annotated metric_function which will be called with a simplified signature.
    """  # noqa: E501 # FIXME CoP
    domain_type = MetricDomainTypes.COLUMN_PAIR
    if issubclass(engine, PandasExecutionEngine):
        if partial_fn_type is None:
            partial_fn_type = MetricPartialFunctionTypes.MAP_SERIES

        partial_fn_type = MetricPartialFunctionTypes(partial_fn_type)
        if partial_fn_type != MetricPartialFunctionTypes.MAP_SERIES:
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                f"""PandasExecutionEngine only supports "{MetricPartialFunctionTypes.MAP_SERIES.value}" for \
"column_pair_function_partial" "partial_fn_type" property."""  # noqa: E501 # FIXME CoP
            )

        def wrapper(metric_fn: Callable):
            assert partial_fn_type is not None  # mypy has trouble type narrowing with closures

            @metric_partial(
                engine=engine,
                partial_fn_type=partial_fn_type,
                domain_type=domain_type,
                **kwargs,
            )
            @wraps(metric_fn)
            def inner_func(  # noqa: PLR0913 # FIXME CoP
                cls,
                execution_engine: PandasExecutionEngine,
                metric_domain_kwargs: dict,
                metric_value_kwargs: dict,
                metrics: Dict[str, Any],
                runtime_configuration: dict,
            ):
                metric_domain_kwargs = get_dbms_compatible_metric_domain_kwargs(
                    metric_domain_kwargs=metric_domain_kwargs,
                    batch_columns_list=metrics["table.columns"],
                )

                (
                    df,
                    compute_domain_kwargs,
                    accessor_domain_kwargs,
                ) = execution_engine.get_compute_domain(
                    domain_kwargs=metric_domain_kwargs, domain_type=domain_type
                )

                # noinspection PyPep8Naming
                column_A_name = accessor_domain_kwargs["column_A"]
                # noinspection PyPep8Naming
                column_B_name = accessor_domain_kwargs["column_B"]

                values = metric_fn(
                    cls,
                    df[column_A_name],
                    df[column_B_name],
                    **metric_value_kwargs,
                    _metrics=metrics,
                )
                return values, compute_domain_kwargs, accessor_domain_kwargs

            return inner_func

        return wrapper

    elif issubclass(engine, SqlAlchemyExecutionEngine):
        if partial_fn_type is None:
            partial_fn_type = MetricPartialFunctionTypes.MAP_FN

        partial_fn_type = MetricPartialFunctionTypes(partial_fn_type)
        if partial_fn_type != MetricPartialFunctionTypes.MAP_FN:
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                f"""SqlAlchemyExecutionEngine only supports "{MetricPartialFunctionTypes.MAP_FN.value}" for \
"column_pair_function_partial" "partial_fn_type" property."""  # noqa: E501 # FIXME CoP
            )

        def wrapper(metric_fn: Callable):
            assert partial_fn_type is not None  # mypy has trouble type narrowing with closures

            @metric_partial(
                engine=engine,
                partial_fn_type=partial_fn_type,
                domain_type=domain_type,
                **kwargs,
            )
            @wraps(metric_fn)
            def inner_func(  # noqa: PLR0913 # FIXME CoP
                cls,
                execution_engine: SqlAlchemyExecutionEngine,
                metric_domain_kwargs: dict,
                metric_value_kwargs: dict,
                metrics: Dict[str, Any],
                runtime_configuration: dict,
            ):
                metric_domain_kwargs = get_dbms_compatible_metric_domain_kwargs(
                    metric_domain_kwargs=metric_domain_kwargs,
                    batch_columns_list=metrics["table.columns"],
                )

                (
                    _selectable,
                    compute_domain_kwargs,
                    accessor_domain_kwargs,
                ) = execution_engine.get_compute_domain(
                    domain_kwargs=metric_domain_kwargs, domain_type=domain_type
                )

                # noinspection PyPep8Naming
                column_A_name = accessor_domain_kwargs["column_A"]
                # noinspection PyPep8Naming
                column_B_name = accessor_domain_kwargs["column_B"]

                column_pair_function = metric_fn(
                    cls,
                    sa.column(column_A_name),
                    sa.column(column_B_name),
                    **metric_value_kwargs,
                    _metrics=metrics,
                )
                return (
                    column_pair_function,
                    compute_domain_kwargs,
                    accessor_domain_kwargs,
                )

            return inner_func

        return wrapper

    elif issubclass(engine, SparkDFExecutionEngine):
        if partial_fn_type is None:
            partial_fn_type = MetricPartialFunctionTypes.MAP_FN

        partial_fn_type = MetricPartialFunctionTypes(partial_fn_type)
        if partial_fn_type != MetricPartialFunctionTypes.MAP_FN:
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                f"""SparkDFExecutionEngine only supports "{MetricPartialFunctionTypes.MAP_FN.value}" for \
"column_pair_function_partial" "partial_fn_type" property."""  # noqa: E501 # FIXME CoP
            )

        def wrapper(metric_fn: Callable):
            @metric_partial(
                engine=engine,
                partial_fn_type=partial_fn_type,
                domain_type=domain_type,
                **kwargs,
            )
            @wraps(metric_fn)
            def inner_func(  # noqa: PLR0913 # FIXME CoP
                cls,
                execution_engine: SparkDFExecutionEngine,
                metric_domain_kwargs: dict,
                metric_value_kwargs: dict,
                metrics: Dict[str, Any],
                runtime_configuration: dict,
            ):
                metric_domain_kwargs = get_dbms_compatible_metric_domain_kwargs(
                    metric_domain_kwargs=metric_domain_kwargs,
                    batch_columns_list=metrics["table.columns"],
                )

                (
                    data,
                    compute_domain_kwargs,
                    accessor_domain_kwargs,
                ) = execution_engine.get_compute_domain(
                    domain_kwargs=metric_domain_kwargs, domain_type=domain_type
                )

                # noinspection PyPep8Naming
                column_A_name = accessor_domain_kwargs["column_A"]
                # noinspection PyPep8Naming
                column_B_name = accessor_domain_kwargs["column_B"]

                column_pair_function = metric_fn(
                    cls,
                    data[column_A_name],
                    data[column_B_name],
                    **metric_value_kwargs,
                    _metrics=metrics,
                )
                return (
                    column_pair_function,
                    compute_domain_kwargs,
                    accessor_domain_kwargs,
                )

            return inner_func

        return wrapper

    else:
        raise ValueError(  # noqa: TRY003, TRY004 # FIXME CoP
            'Unsupported engine for "column_pair_function_partial" metric function decorator.'
        )



from __future__ import annotations

import json
import re
from pprint import pformat as pf
from typing import TYPE_CHECKING

import mistune
import pytest

from great_expectations.checkpoint import UpdateDataDocsAction
from great_expectations.core.expectation_suite import ExpectationSuite
from great_expectations.core.validation_definition import ValidationDefinition
from great_expectations.data_context.util import file_relative_path
from great_expectations.expectations.expectation_configuration import (
    ExpectationConfiguration,
)
from great_expectations.render import RenderedContent, RenderedDocumentContent
from great_expectations.render.renderer import (
    ExpectationSuitePageRenderer,
    ProfilingResultsPageRenderer,
    ValidationResultsPageRenderer,
)

if TYPE_CHECKING:
    from pytest_mock import MockerFixture

    from great_expectations.core.expectation_validation_result import (
        ExpectationValidationResult,
    )

# module level markers
pytestmark = pytest.mark.filesystem


def test_ExpectationSuitePageRenderer_render_expectation_suite_notes():
    result = ExpectationSuitePageRenderer._render_expectation_suite_notes(
        ExpectationSuite(name="test", notes="*hi*")
    )
    # print(RenderedContent.rendered_content_list_to_json(result.text))
    assert RenderedContent.rendered_content_list_to_json(result.text) == [
        "This Expectation suite currently contains 0 total Expectations across 0 columns.",
        {
            "content_block_type": "markdown",
            "markdown": "*hi*",
            "styling": {"parent": {}},
        },
    ]

    result = ExpectationSuitePageRenderer._render_expectation_suite_notes(
        ExpectationSuite(
            name="test",
            notes=["*alpha*", "_bravo_", "charlie"],
        )
    )
    # print(RenderedContent.rendered_content_list_to_json(result.text))
    assert RenderedContent.rendered_content_list_to_json(result.text) == [
        "This Expectation suite currently contains 0 total Expectations across 0 columns.",
        {
            "content_block_type": "markdown",
            "markdown": "*alpha*",
            "styling": {"parent": {}},
        },
        {
            "content_block_type": "markdown",
            "markdown": "_bravo_",
            "styling": {"parent": {}},
        },
        {
            "content_block_type": "markdown",
            "markdown": "charlie",
            "styling": {"parent": {}},
        },
    ]

    result = ExpectationSuitePageRenderer._render_expectation_suite_notes(
        ExpectationSuite(
            name="test",
            notes="*alpha*",
        )
    )
    # print(RenderedContent.rendered_content_list_to_json(result.text))

    try:
        mistune.markdown("*test*")
        assert RenderedContent.rendered_content_list_to_json(result.text) == [
            "This Expectation suite currently contains 0 total Expectations across 0 columns.",
            {
                "content_block_type": "markdown",
                "styling": {"parent": {}},
                "markdown": "*alpha*",
            },
        ]
    except OSError:
        assert RenderedContent.rendered_content_list_to_json(result.text) == [
            "This Expectation suite currently contains 0 total Expectations across 0 columns.",
            "*alpha*",
        ]

    result = ExpectationSuitePageRenderer._render_expectation_suite_notes(
        ExpectationSuite(
            name="test",
            notes=["*alpha*", "_bravo_", "charlie"],
        )
    )
    # print(RenderedContent.rendered_content_list_to_json(result.text))

    try:
        mistune.markdown("*test*")
        assert RenderedContent.rendered_content_list_to_json(result.text) == [
            "This Expectation suite currently contains 0 total Expectations across 0 columns.",
            {
                "content_block_type": "markdown",
                "styling": {"parent": {}},
                "markdown": "*alpha*",
            },
            {
                "content_block_type": "markdown",
                "styling": {"parent": {}},
                "markdown": "_bravo_",
            },
            {
                "content_block_type": "markdown",
                "styling": {"parent": {}},
                "markdown": "charlie",
            },
        ]
    except OSError:
        assert RenderedContent.rendered_content_list_to_json(result.text) == [
            "This Expectation suite currently contains 0 total Expectations across 0 columns.",
            "*alpha*",
            "_bravo_",
            "charlie",
        ]


def test_expectation_summary_in_ExpectationSuitePageRenderer_render_expectation_suite_notes():
    result = ExpectationSuitePageRenderer._render_expectation_suite_notes(
        ExpectationSuite(
            name="test",
            meta={},
            expectations=None,
        )
    )
    # print(RenderedContent.rendered_content_list_to_json(result.text))
    assert RenderedContent.rendered_content_list_to_json(result.text) == [
        "This Expectation suite currently contains 0 total Expectations across 0 columns."
    ]

    result = ExpectationSuitePageRenderer._render_expectation_suite_notes(
        ExpectationSuite(
            name="test",
            notes=["hi"],
        )
    )
    # print(RenderedContent.rendered_content_list_to_json(result.text))

    try:
        mistune.markdown("*test*")
        assert RenderedContent.rendered_content_list_to_json(result.text) == [
            "This Expectation suite currently contains 0 total Expectations across 0 columns.",
            {
                "content_block_type": "markdown",
                "styling": {"parent": {}},
                "markdown": "hi",
            },
        ]
    except OSError:
        assert RenderedContent.rendered_content_list_to_json(result.text) == [
            "This Expectation suite currently contains 0 total Expectations across 0 columns.",
            "hi",
        ]

    result = ExpectationSuitePageRenderer._render_expectation_suite_notes(
        ExpectationSuite(
            name="test",
            meta={},
            expectations=[
                ExpectationConfiguration(
                    type="expect_table_row_count_to_be_between",
                    kwargs={"min_value": 0, "max_value": None},
                ),
                ExpectationConfiguration(type="expect_column_to_exist", kwargs={"column": "x"}),
                ExpectationConfiguration(type="expect_column_to_exist", kwargs={"column": "y"}),
            ],
        )
    )
    # print(RenderedContent.rendered_content_list_to_json(result.text)[0])
    assert (
        RenderedContent.rendered_content_list_to_json(result.text)[0]
        == "This Expectation suite currently contains 3 total Expectations across 2 columns."
    )


def test_ProfilingResultsPageRenderer(
    titanic_profiled_evrs_1: ExpectationValidationResult,
):
    document = ProfilingResultsPageRenderer().render(titanic_profiled_evrs_1)
    assert isinstance(document, RenderedDocumentContent)
    assert len(document.sections) == 8


def test_ValidationResultsPageRenderer_render_validation_header(
    titanic_profiled_evrs_1: ExpectationValidationResult,
):
    validation_header = ValidationResultsPageRenderer._render_validation_header(
        titanic_profiled_evrs_1
    ).to_json_dict()

    expected_validation_header = {
        "content_block_type": "header",
        "styling": {
            "classes": ["col-12", "p-0"],
            "header": {"classes": ["alert", "alert-secondary"]},
        },
        "header": {
            "content_block_type": "string_template",
            "string_template": {
                "template": "Overview",
                "tag": "h5",
                "styling": {"classes": ["m-0"]},
            },
        },
        "subheader": {
            "content_block_type": "string_template",
            "string_template": {
                "template": "${suite_title} ${expectation_suite_name}\n ${data_asset} ${data_asset_name}\n ${status_title} ${html_success_icon} ${success}",  # noqa: E501 # FIXME CoP
                "params": {
                    "suite_title": "Expectation Suite:",
                    "data_asset": "Data asset:",
                    "data_asset_name": None,
                    "status_title": "Status:",
                    "expectation_suite_name": "default",
                    "success": "Failed",
                    "html_success_icon": '<i class="fas fa-times text-danger" aria-hidden="true"></i>',  # noqa: E501 # FIXME CoP
                },
                "styling": {
                    "params": {
                        "suite_title": {"classes": ["h6"]},
                        "status_title": {"classes": ["h6"]},
                        "expectation_suite_name": {
                            "tag": "a",
                            "attributes": {"href": "../../../../expectations/default.html"},
                        },
                    },
                    "classes": ["mb-0", "mt-1"],
                },
            },
        },
    }
    import pprint

    pprint.pprint(validation_header)

    assert validation_header == expected_validation_header


def test_ValidationResultsPageRenderer_render_validation_info(
    titanic_profiled_evrs_1: ExpectationValidationResult,
):
    validation_info = ValidationResultsPageRenderer._render_validation_info(
        titanic_profiled_evrs_1
    ).to_json_dict()
    print(validation_info)

    expected_validation_info = {
        "content_block_type": "table",
        "styling": {
            "classes": ["col-12", "table-responsive", "mt-1"],
            "body": {
                "classes": ["table", "table-sm"],
                "styles": {
                    "margin-bottom": "0.5rem !important",
                    "margin-top": "0.5rem !important",
                },
            },
        },
        "header": {
            "content_block_type": "string_template",
            "string_template": {
                "template": "Info",
                "tag": "h6",
                "styling": {"classes": ["m-0"]},
            },
        },
        "table": [
            ["Great Expectations Version", "0.9.7+17.g02805059.dirty"],
            ["Run Name", "20200322T170247.671855Z"],
            ["Run Time", "2020-03-22T17:02:47Z"],
        ],
    }
    assert validation_info == expected_validation_info


def test_ValidationResultsPageRenderer_render_validation_statistics(
    titanic_profiled_evrs_1,
):
    validation_statistics = ValidationResultsPageRenderer._render_validation_statistics(
        titanic_profiled_evrs_1
    ).to_json_dict()
    print(validation_statistics)
    expected_validation_statistics = {
        "content_block_type": "table",
        "styling": {
            "classes": ["col-6", "table-responsive", "mt-1", "p-1"],
            "body": {
                "classes": ["table", "table-sm"],
                "styles": {
                    "margin-bottom": "0.5rem !important",
                    "margin-top": "0.5rem !important",
                },
            },
        },
        "header": {
            "content_block_type": "string_template",
            "string_template": {
                "template": "Statistics",
                "tag": "h6",
                "styling": {"classes": ["m-0"]},
            },
        },
        "table": [
            ["Evaluated Expectations", 51],
            ["Successful Expectations", 43],
            ["Unsuccessful Expectations", 8],
            ["Success Percent", "≈84.31%"],
        ],
    }

    assert validation_statistics == expected_validation_statistics


@pytest.mark.filesystem
def test_ValidationResultsPageRenderer_render_nested_table_from_dict():
    batch_kwargs = {
        "path": "project_dir/project_path/data/titanic/Titanic.csv",
        "datasource": "Titanic",
        "reader_options": {"sep": None, "engine": "python"},
    }
    batch_kwargs_table = ValidationResultsPageRenderer._render_nested_table_from_dict(
        batch_kwargs, header="Batch Kwargs"
    ).to_json_dict()
    print(batch_kwargs_table)

    expected_batch_kwarg_table = {
        "content_block_type": "table",
        "styling": {
            "body": {
                "classes": ["table", "table-sm"],
                "styles": {
                    "margin-bottom": "0.5rem !important",
                    "margin-top": "0.5rem !important",
                },
            }
        },
        "header": {
            "content_block_type": "string_template",
            "string_template": {
                "template": "Batch Kwargs",
                "tag": "h6",
                "styling": {"classes": ["m-0"]},
            },
        },
        "table": [
            [
                {
                    "content_block_type": "string_template",
                    "styling": {"parent": {"classes": ["pr-3"]}},
                    "string_template": {
                        "template": "$value",
                        "params": {"value": "datasource"},
                        "styling": {"default": {"styles": {"word-break": "break-all"}}},
                    },
                },
                {
                    "content_block_type": "string_template",
                    "styling": {"parent": {"classes": []}},
                    "string_template": {
                        "template": "$value",
                        "params": {"value": "Titanic"},
                        "styling": {"default": {"styles": {"word-break": "break-all"}}},
                    },
                },
            ],
            [
                {
                    "content_block_type": "string_template",
                    "styling": {"parent": {"classes": ["pr-3"]}},
                    "string_template": {
                        "template": "$value",
                        "params": {"value": "path"},
                        "styling": {"default": {"styles": {"word-break": "break-all"}}},
                    },
                },
                {
                    "content_block_type": "string_template",
                    "styling": {"parent": {"classes": []}},
                    "string_template": {
                        "template": "$value",
                        "params": {"value": "project_dir/project_path/data/titanic/Titanic.csv"},
                        "styling": {"default": {"styles": {"word-break": "break-all"}}},
                    },
                },
            ],
            [
                {
                    "content_block_type": "string_template",
                    "styling": {"parent": {"classes": ["pr-3"]}},
                    "string_template": {
                        "template": "$value",
                        "params": {"value": "reader_options"},
                        "styling": {"default": {"styles": {"word-break": "break-all"}}},
                    },
                },
                {
                    "content_block_type": "table",
                    "styling": {
                        "classes": ["col-6", "table-responsive"],
                        "body": {"classes": ["table", "table-sm", "m-0"]},
                        "parent": {"classes": ["pt-0", "pl-0", "border-top-0"]},
                    },
                    "table": [
                        [
                            {
                                "content_block_type": "string_template",
                                "styling": {"parent": {"classes": ["pr-3"]}},
                                "string_template": {
                                    "template": "$value",
                                    "params": {"value": "engine"},
                                    "styling": {"default": {"styles": {"word-break": "break-all"}}},
                                },
                            },
                            {
                                "content_block_type": "string_template",
                                "styling": {"parent": {"classes": []}},
                                "string_template": {
                                    "template": "$value",
                                    "params": {"value": "python"},
                                    "styling": {"default": {"styles": {"word-break": "break-all"}}},
                                },
                            },
                        ],
                        [
                            {
                                "content_block_type": "string_template",
                                "styling": {"parent": {"classes": ["pr-3"]}},
                                "string_template": {
                                    "template": "$value",
                                    "params": {"value": "sep"},
                                    "styling": {"default": {"styles": {"word-break": "break-all"}}},
                                },
                            },
                            {
                                "content_block_type": "string_template",
                                "styling": {"parent": {"classes": []}},
                                "string_template": {
                                    "template": "$value",
                                    "params": {"value": "None"},
                                    "styling": {"default": {"styles": {"word-break": "break-all"}}},
                                },
                            },
                        ],
                    ],
                },
            ],
        ],
    }

    assert batch_kwargs_table == expected_batch_kwarg_table


@pytest.fixture()
def ValidationResultsPageRenderer_render_with_run_info_at_end():
    """
    Rendered validation results with run info at the end
    Returns:
        json string of rendered validation results
    """
    fixture_filename = file_relative_path(
        __file__,
        "./fixtures/ValidationResultsPageRenderer_render_with_run_info_at_end.json",
    )
    with open(fixture_filename) as infile:
        rendered_validation_results = json.load(infile)
        return rendered_validation_results


@pytest.fixture()
def ValidationResultsPageRenderer_render_with_run_info_at_start():
    """
    Rendered validation results with run info at the start
    Returns:
        json string of rendered validation results
    """
    fixture_filename = file_relative_path(
        __file__,
        "./fixtures/ValidationResultsPageRenderer_render_with_run_info_at_start.json",
    )
    with open(fixture_filename) as infile:
        rendered_validation_results = json.load(infile)
        return rendered_validation_results


def test_snapshot_ValidationResultsPageRenderer_render_with_run_info_at_end(
    titanic_profiled_evrs_1: ExpectationValidationResult,
    ValidationResultsPageRenderer_render_with_run_info_at_end,
):
    validation_results_page_renderer = ValidationResultsPageRenderer(run_info_at_end=True)
    rendered_validation_results = validation_results_page_renderer.render(
        titanic_profiled_evrs_1
    ).to_json_dict()

    # replace version of vega-lite in res to match snapshot test
    content_block = rendered_validation_results["sections"][5]["content_blocks"][1]["table"][10][2][
        "content_blocks"
    ][1]
    content_block["graph"]["$schema"] = re.sub(
        r"v\d*\.\d*\.\d*", "v4.8.1", content_block["graph"]["$schema"]
    )
    assert rendered_validation_results == ValidationResultsPageRenderer_render_with_run_info_at_end


def test_snapshot_ValidationResultsPageRenderer_render_with_run_info_at_start(
    titanic_profiled_evrs_1: ExpectationValidationResult,
    ValidationResultsPageRenderer_render_with_run_info_at_start,
):
    validation_results_page_renderer = ValidationResultsPageRenderer(run_info_at_end=False)
    rendered_validation_results = validation_results_page_renderer.render(
        titanic_profiled_evrs_1
    ).to_json_dict()

    # replace version of vega-lite in res to match snapshot test
    content_block = rendered_validation_results["sections"][5]["content_blocks"][1]["table"][10][2][
        "content_blocks"
    ][1]
    content_block["graph"]["$schema"] = re.sub(
        r"v\d*\.\d*\.\d*", "v4.8.1", content_block["graph"]["$schema"]
    )
    assert (
        rendered_validation_results == ValidationResultsPageRenderer_render_with_run_info_at_start
    )


def test_asset_name_is_part_of_resource_info_index(mocker: MockerFixture):
    """
    DefaultSiteIndexBuilder.add_resource_info_to_index_links_dict is what supplies the
    the resource meta-data to the index page.
    This test checks that the asset_name is being provided when checkpoints are run.
    """
    import great_expectations as gx
    from great_expectations.render.renderer.site_builder import DefaultSiteIndexBuilder

    context = gx.get_context(mode="ephemeral")

    add_resource_info_spy = mocker.spy(
        DefaultSiteIndexBuilder,
        "add_resource_info_to_index_links_dict",
    )

    data_asset = context.data_sources.pandas_default.add_csv_asset(
        "my_asset",
        "https://raw.githubusercontent.com/great-expectations/gx_tutorials/main/data/yellow_tripdata_sample_2019-01.csv",
    )

    batch_definition = data_asset.add_batch_definition_whole_dataframe("my_batch")

    # Create Expectation Suite containing two Expectations.
    suite = context.suites.add(ExpectationSuite(name="expectations"))
    suite.add_expectation(
        gx.expectations.ExpectColumnValuesToBeBetween(
            column="passenger_count", min_value=1, max_value=6
        )
    )
    suite.add_expectation(
        gx.expectations.ExpectColumnValuesToBeBetween(column="fare_amount", min_value=0)
    )

    # Create Validation Definition.
    validation_definition = context.validation_definitions.add(
        ValidationDefinition(
            name="validation definition",
            data=batch_definition,
            suite=suite,
        )
    )

    # Create Checkpoint, run Checkpoint, and capture result.
    checkpoint = context.checkpoints.add(
        gx.checkpoint.checkpoint.Checkpoint(
            name="checkpoint",
            validation_definitions=[validation_definition],
            actions=[UpdateDataDocsAction(name="update_data_docs")],
        )
    )

    checkpoint.run()

    # check `DefaultSiteIndexBuilder.add_resource_info_to_index_links_dict` has the asset name
    add_resource_info_spy.assert_called()
    last_call = add_resource_info_spy.call_args_list[-1]
    print(f"Last call kwargs:\n{pf(last_call.kwargs, depth=1)}")
    assert last_call.kwargs["asset_name"] == data_asset.name



from typing import Optional

from great_expectations.compatibility.pydantic import StrictStr
from great_expectations.expectations.model_field_types import ConditionParser
from great_expectations.metrics.metric import Metric, NonEmptyString, _MetricResult


class ColumnMetric(Metric[_MetricResult], kw_only=True):
    column: NonEmptyString
    row_condition: Optional[StrictStr] = None
    condition_parser: Optional[ConditionParser] = None



import pandas as pd
import pytest

import great_expectations.expectations as gxe
from great_expectations.data_context.data_context.abstract_data_context import AbstractDataContext


class ExpectColumnValuesAsStringToBePositiveInteger(gxe.ExpectColumnValuesToMatchRegex):
    regex: str = "^\\d+$"


@pytest.mark.big
def test_expect_column_values_as_string_to_be_positive_integers_pass(
    empty_data_context: AbstractDataContext,
):
    df = pd.DataFrame({"a": ["1", "2", "3", "4", "5"]})
    data_asset = empty_data_context.data_sources.pandas_default.add_dataframe_asset("my_dataframe")
    batch = data_asset.add_batch_definition_whole_dataframe("my_batch_definition").get_batch(
        batch_parameters={"dataframe": df}
    )

    result = batch.validate(ExpectColumnValuesAsStringToBePositiveInteger(column="a"))

    assert result.success


@pytest.mark.big
def test_expect_column_values_as_string_to_be_positive_integers_fail(
    empty_data_context: AbstractDataContext,
):
    df = pd.DataFrame({"a": ["1", "2", "3", "4", "a"]})

    data_asset = empty_data_context.data_sources.pandas_default.add_dataframe_asset("my_dataframe")
    batch = data_asset.add_batch_definition_whole_dataframe("my_batch_definition").get_batch(
        batch_parameters={"dataframe": df}
    )

    result = batch.validate(ExpectColumnValuesAsStringToBePositiveInteger(column="a"))

    assert not result.success



from __future__ import annotations

import logging
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Sequence,
    Tuple,
    Union,
)

import great_expectations.exceptions as gx_exceptions
from great_expectations.expectations.metrics.util import (
    MAX_RESULT_RECORDS,
    get_dbms_compatible_metric_domain_kwargs,
)

if TYPE_CHECKING:
    from great_expectations.execution_engine import (
        PandasExecutionEngine,
        SparkDFExecutionEngine,
        SqlAlchemyExecutionEngine,
    )

from great_expectations.compatibility.pyspark import functions as F
from great_expectations.compatibility.sqlalchemy import sqlalchemy as sa
from great_expectations.execution_engine.sqlalchemy_dialect import GXSqlDialect
from great_expectations.expectations.metrics.map_metric_provider.is_sqlalchemy_metric_selectable import (  # noqa: E501 # FIXME CoP
    _is_sqlalchemy_metric_selectable,
)

if TYPE_CHECKING:
    import pandas as pd

    from great_expectations.compatibility import pyspark, sqlalchemy


logger = logging.getLogger(__name__)


def _pandas_column_map_condition_values(
    cls,
    execution_engine: PandasExecutionEngine,
    metric_domain_kwargs: dict,
    metric_value_kwargs: dict,
    metrics: Dict[str, Any],
    **kwargs,
) -> list[dict]:
    """Return values from the specified domain that match the map-style metric in the metrics dictionary."""  # noqa: E501 # FIXME CoP
    (
        boolean_mapped_unexpected_values,
        compute_domain_kwargs,
        accessor_domain_kwargs,
    ) = metrics["unexpected_condition"]

    if "column" not in accessor_domain_kwargs:
        raise ValueError(  # noqa: TRY003 # FIXME CoP
            """No "column" found in provided metric_domain_kwargs, but it is required for a column map metric
(_pandas_column_map_condition_values).
"""  # noqa: E501 # FIXME CoP
        )

    accessor_domain_kwargs = get_dbms_compatible_metric_domain_kwargs(
        metric_domain_kwargs=accessor_domain_kwargs,
        batch_columns_list=metrics["table.columns"],
    )

    column_name: Union[str, sqlalchemy.quoted_name] = accessor_domain_kwargs["column"]

    df = execution_engine.get_domain_records(domain_kwargs=compute_domain_kwargs)

    ###
    # NOTE: 20201111 - JPC - in the map_series / map_condition_series world (pandas), we
    # currently handle filter_column_isnull differently than other map_fn / map_condition
    # cases.
    ###
    filter_column_isnull = kwargs.get(
        "filter_column_isnull", getattr(cls, "filter_column_isnull", False)
    )
    if filter_column_isnull:
        df = df[df[column_name].notnull()]

    domain_values = df[column_name]

    domain_values = domain_values[
        boolean_mapped_unexpected_values == True  # noqa: E712 # FIXME CoP
    ]

    result_format = metric_value_kwargs["result_format"]

    if result_format["result_format"] == "COMPLETE":
        return list(domain_values[:MAX_RESULT_RECORDS])

    limit = min(result_format["partial_unexpected_count"], MAX_RESULT_RECORDS)
    return list(domain_values[:limit])


def _pandas_column_map_condition_value_counts(
    cls,
    execution_engine: PandasExecutionEngine,
    metric_domain_kwargs: dict,
    metric_value_kwargs: dict,
    metrics: Dict[str, Any],
    **kwargs,
) -> pd.Series[int]:
    """Returns respective value counts for distinct column values"""
    (
        boolean_mapped_unexpected_values,
        compute_domain_kwargs,
        accessor_domain_kwargs,
    ) = metrics["unexpected_condition"]

    if "column" not in accessor_domain_kwargs:
        raise ValueError(  # noqa: TRY003 # FIXME CoP
            """No "column" found in provided metric_domain_kwargs, but it is required for a column map metric
(_pandas_column_map_condition_value_counts).
"""  # noqa: E501 # FIXME CoP
        )

    accessor_domain_kwargs = get_dbms_compatible_metric_domain_kwargs(
        metric_domain_kwargs=accessor_domain_kwargs,
        batch_columns_list=metrics["table.columns"],
    )

    column_name: Union[str, sqlalchemy.quoted_name] = accessor_domain_kwargs["column"]

    df = execution_engine.get_domain_records(domain_kwargs=compute_domain_kwargs)

    ###
    # NOTE: 20201111 - JPC - in the map_series / map_condition_series world (pandas), we
    # currently handle filter_column_isnull differently than other map_fn / map_condition
    # cases.
    ###
    filter_column_isnull = kwargs.get(
        "filter_column_isnull", getattr(cls, "filter_column_isnull", False)
    )
    if filter_column_isnull:
        df = df[df[column_name].notnull()]

    domain_values = df[column_name]

    result_format = metric_value_kwargs["result_format"]
    value_counts = None
    try:
        value_counts = domain_values[boolean_mapped_unexpected_values].value_counts()
    except ValueError:
        try:
            value_counts = (
                domain_values[boolean_mapped_unexpected_values].apply(tuple).value_counts()
            )
        except ValueError:
            pass

    if not value_counts:
        raise gx_exceptions.MetricComputationError("Unable to compute value counts")  # noqa: TRY003 # FIXME CoP

    if result_format["result_format"] == "COMPLETE":
        return value_counts

    return value_counts[result_format["partial_unexpected_count"]]


def _sqlalchemy_column_map_condition_values(
    cls,
    execution_engine: SqlAlchemyExecutionEngine,
    metric_domain_kwargs: dict,
    metric_value_kwargs: dict,
    metrics: Dict[str, Tuple],
    **kwargs,
) -> list[dict]:
    """
    Particularly for the purpose of finding unexpected values, returns all the metric values which do not meet an
    expected Expectation condition for ColumnMapExpectation Expectations.
    """  # noqa: E501 # FIXME CoP
    unexpected_condition, compute_domain_kwargs, accessor_domain_kwargs = metrics[
        "unexpected_condition"
    ]

    if "column" not in accessor_domain_kwargs:
        raise ValueError(  # noqa: TRY003 # FIXME CoP
            """No "column" found in provided metric_domain_kwargs, but it is required for a column map metric
(_sqlalchemy_column_map_condition_values).
"""  # noqa: E501 # FIXME CoP
        )

    accessor_domain_kwargs = get_dbms_compatible_metric_domain_kwargs(
        metric_domain_kwargs=accessor_domain_kwargs,
        batch_columns_list=metrics["table.columns"],
    )

    column_name: Union[str, sqlalchemy.quoted_name] = accessor_domain_kwargs["column"]

    selectable = execution_engine.get_domain_records(domain_kwargs=compute_domain_kwargs)

    query = sa.select(sa.column(column_name).label("unexpected_values")).where(unexpected_condition)  # type: ignore[var-annotated] # FIXME CoP
    if not _is_sqlalchemy_metric_selectable(map_metric_provider=cls):
        if hasattr(selectable, "subquery"):
            query = query.select_from(selectable.subquery())
        else:
            query = query.select_from(selectable)  # type: ignore[arg-type] # FIXME CoP

    result_format = metric_value_kwargs["result_format"]

    if result_format["result_format"] != "COMPLETE":
        query = query.limit(result_format["partial_unexpected_count"])
    elif (
        result_format["result_format"] == "COMPLETE"
        and execution_engine.engine.dialect.name.lower() == GXSqlDialect.BIGQUERY
    ):
        logger.warning(
            "BigQuery imposes a limit of 10000 parameters on individual queries; "
            "if your data contains more than 10000 columns your results will be truncated."
        )
        query = query.limit(10000)  # BigQuery upper bound on query parameters

    return [
        val.unexpected_values
        for val in execution_engine.execute_query(query).fetchmany(MAX_RESULT_RECORDS)
    ]


def _sqlalchemy_column_map_condition_value_counts(
    cls,
    execution_engine: SqlAlchemyExecutionEngine,
    metric_domain_kwargs: dict,
    metric_value_kwargs: dict,
    metrics: Dict[str, Any],
    **kwargs,
) -> Union[Sequence[sa.Row[Any]], Any]:
    """
    Returns value counts for all the metric values which do not meet an expected Expectation condition for instances
    of ColumnMapExpectation.
    """  # noqa: E501 # FIXME CoP
    unexpected_condition, compute_domain_kwargs, accessor_domain_kwargs = metrics[
        "unexpected_condition"
    ]

    if "column" not in accessor_domain_kwargs:
        raise ValueError(  # noqa: TRY003 # FIXME CoP
            """No "column" found in provided metric_domain_kwargs, but it is required for a column map metric
(_sqlalchemy_column_map_condition_value_counts).
"""  # noqa: E501 # FIXME CoP
        )

    accessor_domain_kwargs = get_dbms_compatible_metric_domain_kwargs(
        metric_domain_kwargs=accessor_domain_kwargs,
        batch_columns_list=metrics["table.columns"],
    )

    column_name: Union[str, sqlalchemy.quoted_name] = accessor_domain_kwargs["column"]

    column: sa.Column = sa.column(column_name)  # type: ignore[assignment] # FIXME CoP

    selectable = execution_engine.get_domain_records(domain_kwargs=compute_domain_kwargs)

    query = sa.select(column, sa.func.count(column)).where(unexpected_condition).group_by(column)
    if not _is_sqlalchemy_metric_selectable(map_metric_provider=cls):
        query = query.select_from(selectable)  # type: ignore[arg-type] # FIXME CoP

    return execution_engine.execute_query(query).fetchall()


def _spark_column_map_condition_values(
    cls,
    execution_engine: SparkDFExecutionEngine,
    metric_domain_kwargs: dict,
    metric_value_kwargs: dict,
    metrics: Dict[str, Any],
    **kwargs,
) -> list[dict]:
    """Return values from the specified domain that match the map-style metric in the metrics dictionary."""  # noqa: E501 # FIXME CoP
    unexpected_condition, compute_domain_kwargs, accessor_domain_kwargs = metrics[
        "unexpected_condition"
    ]

    if "column" not in accessor_domain_kwargs:
        raise ValueError(  # noqa: TRY003 # FIXME CoP
            """No "column" found in provided metric_domain_kwargs, but it is required for a column map metric
(_spark_column_map_condition_values).
"""  # noqa: E501 # FIXME CoP
        )

    accessor_domain_kwargs = get_dbms_compatible_metric_domain_kwargs(
        metric_domain_kwargs=accessor_domain_kwargs,
        batch_columns_list=metrics["table.columns"],
    )

    column_name: Union[str, sqlalchemy.quoted_name] = accessor_domain_kwargs["column"]

    df = execution_engine.get_domain_records(domain_kwargs=compute_domain_kwargs)

    # withColumn is required to transform window functions returned by some metrics to boolean mask
    data = df.withColumn("__unexpected", unexpected_condition)
    filtered = data.filter(F.col("__unexpected") == True).drop(  # noqa: E712 # FIXME CoP
        F.col("__unexpected")
    )

    result_format = metric_value_kwargs["result_format"]

    # note that without an explicit column alias,
    # spark will use only the final portion
    # of a nested column as the column name
    if result_format["result_format"] == "COMPLETE":
        query = filtered.select(F.col(column_name).alias(column_name)).limit(MAX_RESULT_RECORDS)
    else:
        limit = min(result_format["partial_unexpected_count"], MAX_RESULT_RECORDS)
        query = filtered.select(F.col(column_name).alias(column_name)).limit(limit)
    return [row[column_name] for row in query.collect()]


def _spark_column_map_condition_value_counts(
    cls,
    execution_engine: SparkDFExecutionEngine,
    metric_domain_kwargs: dict,
    metric_value_kwargs: dict,
    metrics: Dict[str, Any],
    **kwargs,
) -> list[pyspark.Row]:
    unexpected_condition, compute_domain_kwargs, accessor_domain_kwargs = metrics[
        "unexpected_condition"
    ]

    if "column" not in accessor_domain_kwargs:
        raise ValueError(  # noqa: TRY003 # FIXME CoP
            """No "column" found in provided metric_domain_kwargs, but it is required for a column map metric
(_spark_column_map_condition_value_counts).
"""  # noqa: E501 # FIXME CoP
        )

    accessor_domain_kwargs = get_dbms_compatible_metric_domain_kwargs(
        metric_domain_kwargs=accessor_domain_kwargs,
        batch_columns_list=metrics["table.columns"],
    )

    column_name: Union[str, sqlalchemy.quoted_name] = accessor_domain_kwargs["column"]

    df: pyspark.DataFrame = execution_engine.get_domain_records(domain_kwargs=compute_domain_kwargs)

    # withColumn is required to transform window functions returned by some metrics to boolean mask
    data = df.withColumn("__unexpected", unexpected_condition)
    filtered = data.filter(F.col("__unexpected") == True).drop(  # noqa: E712 # FIXME CoP
        F.col("__unexpected")
    )

    result_format = metric_value_kwargs["result_format"]

    value_counts = filtered.groupBy(F.col(column_name).alias(column_name)).count()
    if result_format["result_format"] == "COMPLETE":
        rows = value_counts.collect()
    else:
        rows = value_counts.collect()[: result_format["partial_unexpected_count"]]
    return rows



from __future__ import annotations

import json
import logging
from pprint import pformat as pf

import requests
from requests.adapters import HTTPAdapter, Retry

from great_expectations import __version__
from great_expectations.compatibility.typing_extensions import override

DEFAULT_TIMEOUT = 20

LOGGER = logging.getLogger(__name__)


def _log_request_method_and_response(r: requests.Response, *args, **kwargs):
    LOGGER.info(f"{r.request.method} {r.request.url} - {r}")
    try:
        LOGGER.debug(f"{r}\n{pf(r.json(), depth=3)}")
    except json.JSONDecodeError:
        LOGGER.debug(f"{r}\n{r.content.decode()}")
    except Exception as other_err:
        LOGGER.info(f"{r} - Error logging response {other_err!r}")


class _TimeoutHTTPAdapter(HTTPAdapter):
    # https://stackoverflow.com/a/62044100
    # Session-wide timeouts are not supported by requests
    # but are discussed in detail here: https://github.com/psf/requests/issues/3070
    def __init__(self, *args, **kwargs) -> None:
        self.timeout = kwargs.pop("timeout", DEFAULT_TIMEOUT)
        super().__init__(*args, **kwargs)

    @override
    def send(self, request: requests.PreparedRequest, **kwargs) -> requests.Response:  # type: ignore[override] # FIXME CoP
        kwargs["timeout"] = kwargs.get("timeout", self.timeout)
        return super().send(request, **kwargs)


def create_session(
    access_token: str,
    retry_count: int = 5,
    backoff_factor: float = 1.0,
    timeout: int = DEFAULT_TIMEOUT,
) -> requests.Session:
    session = requests.Session()
    session = _update_headers(session=session, access_token=access_token)
    session = _mount_adapter(
        session=session,
        timeout=timeout,
        retry_count=retry_count,
        backoff_factor=backoff_factor,
    )
    # add an event hook to log outgoing http requests
    # https://requests.readthedocs.io/en/latest/user/advanced/#event-hooks
    session.hooks["response"].append(_log_request_method_and_response)
    return session


def _update_headers(session: requests.Session, access_token: str) -> requests.Session:
    headers = {
        "Content-Type": "application/vnd.api+json",
        "Authorization": f"Bearer {access_token}",
        "Gx-Version": __version__,
    }
    session.headers.update(headers)
    return session


def _mount_adapter(
    session: requests.Session, timeout: int, retry_count: int, backoff_factor: float
) -> requests.Session:
    retries = Retry(total=retry_count, backoff_factor=backoff_factor)
    adapter = _TimeoutHTTPAdapter(timeout=timeout, max_retries=retries)
    for protocol in ("http://", "https://"):
        session.mount(protocol, adapter)
    return session



from unittest import mock

import pandas as pd
import pytest

from great_expectations.compatibility.sqlalchemy import sqlalchemy as sa
from great_expectations.compatibility.sqlalchemy_compatibility_wrappers import (
    add_dataframe_to_db,
)
from great_expectations.execution_engine.sqlalchemy_batch_data import (
    SqlAlchemyBatchData,
)
from great_expectations.execution_engine.sqlalchemy_execution_engine import (
    SqlAlchemyExecutionEngine,
)
from great_expectations.expectations.metrics.table_metrics.table_head import TableHead
from tests.test_utils import get_sqlite_temp_table_names_from_engine


@pytest.fixture
def sqlite_engine():
    sqlite_engine = sa.create_engine("sqlite://")
    dataframe = pd.DataFrame({"a": [1, 2, 3, 4], "b": [2, 3, 4, 5]})
    add_dataframe_to_db(
        df=dataframe,
        name="test_table",
        con=sqlite_engine,
        index=False,
    )
    return sqlite_engine


@pytest.fixture
def sqlite_batch_with_table_name(sqlite_engine) -> SqlAlchemyExecutionEngine:
    execution_engine: SqlAlchemyExecutionEngine = SqlAlchemyExecutionEngine(engine=sqlite_engine)
    batch_data = SqlAlchemyBatchData(
        execution_engine=execution_engine,
        table_name="test_table",
    )
    execution_engine.load_batch_data("__", batch_data)
    return execution_engine


@pytest.fixture
def sqlite_batch_with_selectable_with_temp_table(
    sqlite_engine,
) -> SqlAlchemyExecutionEngine:
    execution_engine: SqlAlchemyExecutionEngine = SqlAlchemyExecutionEngine(engine=sqlite_engine)
    selectable = sa.select("*").select_from(sa.text("main.test_table"))
    batch_data = SqlAlchemyBatchData(
        execution_engine=execution_engine, selectable=selectable, create_temp_table=True
    )
    execution_engine.load_batch_data("__", batch_data)
    return execution_engine


@pytest.fixture
def sqlite_batch_with_selectable_without_temp_table(
    sqlite_engine,
) -> SqlAlchemyExecutionEngine:
    execution_engine: SqlAlchemyExecutionEngine = SqlAlchemyExecutionEngine(engine=sqlite_engine)
    selectable = sa.select("*").select_from(sa.text("main.test_table"))
    batch_data = SqlAlchemyBatchData(
        execution_engine=execution_engine,
        selectable=selectable,
        create_temp_table=False,
    )
    execution_engine.load_batch_data("__", batch_data)
    return execution_engine


@pytest.mark.sqlite
@pytest.mark.parametrize(
    "execution_engine, n_rows, fetch_all, expected_shape, expected_columns, expected_values, expected_temp_tables",  # noqa: E501 # FIXME CoP
    [
        (
            "sqlite_batch_with_table_name",
            2,
            False,
            (2, 2),
            ["a", "b"],
            [[1, 2], [2, 3]],
            0,
        ),
        (
            "sqlite_batch_with_table_name",
            None,
            True,
            (4, 2),
            ["a", "b"],
            [[1, 2], [2, 3], [3, 4], [4, 5]],
            0,
        ),
        (
            "sqlite_batch_with_selectable_with_temp_table",
            2,
            False,
            (2, 2),
            ["a", "b"],
            [[1, 2], [2, 3]],
            1,
        ),
        (
            "sqlite_batch_with_selectable_with_temp_table",
            None,
            True,
            (4, 2),
            ["a", "b"],
            [[1, 2], [2, 3], [3, 4], [4, 5]],
            1,
        ),
        (
            "sqlite_batch_with_selectable_without_temp_table",
            2,
            False,
            (2, 2),
            ["a", "b"],
            [[1, 2], [2, 3]],
            0,
        ),
        (
            "sqlite_batch_with_selectable_without_temp_table",
            None,
            True,
            (4, 2),
            ["a", "b"],
            [[1, 2], [2, 3], [3, 4], [4, 5]],
            0,
        ),
    ],
)
def test_table_head_sqlite(
    execution_engine,
    n_rows,
    fetch_all,
    expected_shape,
    expected_columns,
    expected_values,
    request,
    expected_temp_tables,
):
    engine = request.getfixturevalue(execution_engine)
    table_head = TableHead()
    res = table_head._sqlalchemy(
        execution_engine=engine,
        metric_domain_kwargs={},
        metric_value_kwargs={"n_rows": n_rows, "fetch_all": fetch_all},
        metrics={},
        runtime_configuration={},
    )

    assert res.shape == expected_shape
    assert res.columns.tolist() == expected_columns
    assert res.values.tolist() == expected_values
    assert len(get_sqlite_temp_table_names_from_engine(engine.engine)) == expected_temp_tables


@pytest.mark.sqlite
@pytest.mark.parametrize(
    "execution_engine",
    [
        "sqlite_batch_with_table_name",
        "sqlite_batch_with_selectable_with_temp_table",
        "sqlite_batch_with_selectable_without_temp_table",
    ],
)
@pytest.mark.parametrize(
    "n_rows",
    [None, 0, 1, 2],
)
@pytest.mark.parametrize(
    "fetch_all",
    [None, True, False],
)
def test_limit_included_in_head_query(
    execution_engine,
    n_rows,
    fetch_all,
    request,
):
    engine = request.getfixturevalue(execution_engine)
    table_head = TableHead()

    with mock.patch(
        "great_expectations.compatibility.sqlalchemy_and_pandas.pd.read_sql"
    ) as mock_node:
        table_head._sqlalchemy(
            execution_engine=engine,
            metric_domain_kwargs={},
            metric_value_kwargs={"n_rows": n_rows, "fetch_all": fetch_all},
            metrics={},
            runtime_configuration={},
        )

        _args, kwargs = mock_node.call_args
        mock_node.assert_called_once()

        assert ("limit" in str(kwargs["sql"]).lower()) == (fetch_all is not True)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Dict, Iterator, List, cast
from unittest import mock

import pytest

import great_expectations.exceptions as ge_exceptions
from great_expectations.compatibility import azure
from great_expectations.datasource.fluent import PandasAzureBlobStorageDatasource
from great_expectations.datasource.fluent.config_str import ConfigStr
from great_expectations.datasource.fluent.data_asset.path.pandas.generated_assets import CSVAsset
from great_expectations.datasource.fluent.dynamic_pandas import PANDAS_VERSION
from great_expectations.datasource.fluent.pandas_azure_blob_storage_datasource import (
    PandasAzureBlobStorageDatasourceError,
)

if TYPE_CHECKING:
    from great_expectations.data_context.data_context.file_data_context import (
        FileDataContext,
    )

logger = logging.getLogger(__file__)


if not (azure.storage and azure.BlobServiceClient and azure.ContainerClient):  # type: ignore[truthy-function] # False if NotImported
    pytest.skip(
        'Could not import "azure.storage.blob" from Microsoft Azure cloud',
        allow_module_level=True,
    )


# apply markers to entire test module
pytestmark = [
    pytest.mark.skipif(
        PANDAS_VERSION < 1.2, reason=f"Fluent pandas not supported on {PANDAS_VERSION}"
    )
]


class MockContainerClient:
    def walk_blobs(
        self,
        name_starts_with: str | None = None,
        include: Any | None = None,
        delimiter: str = "/",
        **kwargs,
    ) -> Iterator:
        return iter([])


class MockBlobServiceClient:
    def get_container_client(self, container: str) -> azure.ContainerClient:
        return cast("azure.ContainerClient", MockContainerClient())


def _build_pandas_abs_datasource(
    azure_options: Dict[str, Any] | None = None,
) -> PandasAzureBlobStorageDatasource:
    azure_client: azure.BlobServiceClient = cast("azure.BlobServiceClient", MockBlobServiceClient())
    pandas_abs_datasource = PandasAzureBlobStorageDatasource(
        name="pandas_abs_datasource",
        azure_options=azure_options or {},
    )
    pandas_abs_datasource._azure_client = azure_client
    return pandas_abs_datasource


@pytest.fixture
def pandas_abs_datasource() -> PandasAzureBlobStorageDatasource:
    pandas_abs_datasource: PandasAzureBlobStorageDatasource = _build_pandas_abs_datasource()
    return pandas_abs_datasource


@pytest.fixture
def object_keys() -> List[str]:
    return [
        "yellow_tripdata_sample_2024-01.csv",
        "yellow_tripdata_sample_2024-02.csv",
        "yellow_tripdata_sample_2024-03.csv",
        "yellow_tripdata_sample_2024-04.csv",
        "yellow_tripdata_sample_2024-05.csv",
        "yellow_tripdata_sample_2024-06.csv",
        "yellow_tripdata_sample_2024-07.csv",
        "yellow_tripdata_sample_2024-08.csv",
        "yellow_tripdata_sample_2024-09.csv",
        "yellow_tripdata_sample_2024-10.csv",
        "yellow_tripdata_sample_2024-11.csv",
        "yellow_tripdata_sample_2024-12.csv",
    ]


@pytest.mark.big
def test_construct_pandas_abs_datasource_with_account_url_and_credential():
    pandas_abs_datasource = PandasAzureBlobStorageDatasource(
        name="pandas_abs_datasource",
        azure_options={
            "account_url": "my_account_url.blob.core.windows.net",
            "credential": "my_credential",
        },
    )
    azure_client: azure.BlobServiceClient = pandas_abs_datasource._get_azure_client()
    assert azure_client is not None
    assert pandas_abs_datasource.name == "pandas_abs_datasource"


@pytest.mark.big
def test_construct_pandas_abs_datasource_with_account_url_and_config_credential(
    monkeypatch: pytest.MonkeyPatch, empty_file_context: FileDataContext
):
    monkeypatch.setenv("MY_CRED", "my_secret_credential")

    pandas_abs_datasource = PandasAzureBlobStorageDatasource(
        name="pandas_abs_datasource",
        azure_options={
            "account_url": "my_account_url.blob.core.windows.net",
            "credential": r"${MY_CRED}",
        },
    )

    # attach data_context to enable config substitution
    pandas_abs_datasource._data_context = empty_file_context

    credential = pandas_abs_datasource.azure_options["credential"]
    assert isinstance(credential, ConfigStr)
    assert (
        credential.get_config_value(config_provider=empty_file_context.config_provider)
        == "my_secret_credential"
    )

    azure_client: azure.BlobServiceClient = pandas_abs_datasource._get_azure_client()
    assert azure_client is not None
    assert pandas_abs_datasource.name == "pandas_abs_datasource"


@pytest.mark.big
def test_construct_pandas_abs_datasource_with_conn_str_and_credential():
    pandas_abs_datasource = PandasAzureBlobStorageDatasource(
        name="pandas_abs_datasource",
        azure_options={  # Representative of format noted in official docs
            "conn_str": "DefaultEndpointsProtocol=https;AccountName=storagesample;AccountKey=my_account_key",  # noqa: E501 # FIXME CoP
            "credential": "my_credential",
        },
    )
    azure_client: azure.BlobServiceClient = pandas_abs_datasource._get_azure_client()
    assert azure_client is not None
    assert pandas_abs_datasource.name == "pandas_abs_datasource"


@pytest.mark.big
def test_construct_pandas_abs_datasource_with_valid_account_url_assigns_account_name():
    pandas_abs_datasource = PandasAzureBlobStorageDatasource(
        name="pandas_abs_datasource",
        azure_options={
            "account_url": "my_account_url.blob.core.windows.net",
            "credential": "my_credential",
        },
    )
    azure_client: azure.BlobServiceClient = pandas_abs_datasource._get_azure_client()
    assert azure_client is not None
    assert pandas_abs_datasource.name == "pandas_abs_datasource"


@pytest.mark.big
def test_construct_pandas_abs_datasource_with_valid_conn_str_assigns_account_name():
    pandas_abs_datasource = PandasAzureBlobStorageDatasource(
        name="pandas_abs_datasource",
        azure_options={  # Representative of format noted in official docs
            "conn_str": "DefaultEndpointsProtocol=https;AccountName=storagesample;AccountKey=my_account_key",  # noqa: E501 # FIXME CoP
            "credential": "my_credential",
        },
    )
    azure_client: azure.BlobServiceClient = pandas_abs_datasource._get_azure_client()
    assert azure_client is not None
    assert pandas_abs_datasource.name == "pandas_abs_datasource"


@pytest.mark.big
def test_construct_pandas_abs_datasource_with_multiple_auth_methods_raises_error():
    # Raises error in DataContext's schema validation due to having both `account_url` and `conn_str`  # noqa: E501 # FIXME CoP
    with pytest.raises(PandasAzureBlobStorageDatasourceError):
        pandas_abs_datasource = PandasAzureBlobStorageDatasource(
            name="pandas_abs_datasource",
            azure_options={
                "account_url": "account.blob.core.windows.net",
                "conn_str": "DefaultEndpointsProtocol=https;AccountName=storagesample;AccountKey=my_account_key",  # noqa: E501 # FIXME CoP
                "credential": "my_credential",
            },
        )
        _ = pandas_abs_datasource._get_azure_client()


@pytest.mark.big
@mock.patch(
    "great_expectations.datasource.fluent.data_connector.azure_blob_storage_data_connector.list_azure_keys"
)
@mock.patch("azure.storage.blob.BlobServiceClient")
def test_add_csv_asset_to_datasource(
    mock_azure_client,
    mock_list_keys,
    object_keys: List[str],
    pandas_abs_datasource: PandasAzureBlobStorageDatasource,
):
    mock_list_keys.return_value = object_keys
    asset = pandas_abs_datasource.add_csv_asset(
        name="csv_asset",
        abs_container="my_container",
    )
    assert asset.name == "csv_asset"


@pytest.mark.big
@mock.patch(
    "great_expectations.datasource.fluent.data_connector.azure_blob_storage_data_connector.list_azure_keys"
)
@mock.patch("azure.storage.blob.BlobServiceClient")
def test_construct_csv_asset_directly(mock_azure_client, mock_list_keys, object_keys: List[str]):
    mock_list_keys.return_value = object_keys
    asset = CSVAsset(  # type: ignore[call-arg] # FIXME CoP
        name="csv_asset",
    )
    assert asset.name == "csv_asset"


@pytest.mark.big
@mock.patch(
    "great_expectations.datasource.fluent.data_connector.azure_blob_storage_data_connector.list_azure_keys"
)
@mock.patch("azure.storage.blob.BlobServiceClient")
def test_csv_asset_with_batching_regex_named_parameters(
    mock_azure_client,
    mock_list_keys,
    object_keys: List[str],
    pandas_abs_datasource: PandasAzureBlobStorageDatasource,
):
    mock_list_keys.return_value = object_keys
    regex = r"yellow_tripdata_sample_(?P<year>\d{4})-(?P<month>\d{2})\.csv"
    asset = pandas_abs_datasource.add_csv_asset(
        name="csv_asset",
        abs_container="my_container",
    )
    batch_def = asset.add_batch_definition_monthly(name="batch def", regex=regex)
    options = asset.get_batch_parameters_keys(partitioner=batch_def.partitioner)
    assert options == ("path", "year", "month")


@pytest.mark.big
@mock.patch(
    "great_expectations.datasource.fluent.data_connector.azure_blob_storage_data_connector.list_azure_keys"
)
@mock.patch("azure.storage.blob.BlobServiceClient")
def test_csv_asset_with_non_string_batching_regex_named_parameters(
    mock_azure_client,
    mock_list_keys,
    object_keys: List[str],
    pandas_abs_datasource: PandasAzureBlobStorageDatasource,
):
    mock_list_keys.return_value = object_keys
    asset = pandas_abs_datasource.add_csv_asset(
        name="csv_asset",
        abs_container="my_container",
    )
    with pytest.raises(ge_exceptions.InvalidBatchRequestError):
        # price is an int which will raise an error
        asset.build_batch_request({"name": "alex", "timestamp": "1234567890", "price": 1300})


@pytest.mark.big
@mock.patch(
    "great_expectations.datasource.fluent.data_connector.azure_blob_storage_data_connector.list_azure_keys"
)
@mock.patch("azure.storage.blob.BlobServiceClient")
def test_add_csv_asset_with_recursive_file_discovery_to_datasource(
    mock_azure_client,
    mock_list_keys,
    object_keys: List[str],
    pandas_abs_datasource: PandasAzureBlobStorageDatasource,
):
    """
    Tests that the abs_recursive_file_discovery-flag is passed on
    to the list_keys-function as the recursive-parameter

    This makes the list_keys-function search and return files also
    from sub-directories on Azure, not just the files in the folder
    specified with the abs_name_starts_with-parameter
    """
    mock_list_keys.return_value = object_keys
    pandas_abs_datasource.add_csv_asset(
        name="csv_asset",
        abs_container="my_container",
        abs_recursive_file_discovery=True,
    )
    assert "recursive" in mock_list_keys.call_args.kwargs
    assert mock_list_keys.call_args.kwargs["recursive"] is True



import pandas as pd
import pytest

from great_expectations.datasource.fluent.interfaces import Batch
from great_expectations.metrics.multi_column.sum_equal import (
    MultiColumnSumEqualUnexpectedCount,
    MultiColumnSumEqualUnexpectedCountResult,
)
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.metrics.conftest import (
    PANDAS_DATA_SOURCES,
    SPARK_DATA_SOURCES,
    SQL_DATA_SOURCES,
)

COL_A = "A"
COL_B = "B"
COL_A_WITH_NULLS = "A_WITH_NULLS"
COL_B_WITH_NULLS = "B_WITH_NULLS"
DATA_FRAME = pd.DataFrame(
    {
        COL_A: [1, 2, 3, 4],
        COL_B: [4, 3, 2, 1],
    },
)
SUCCESS_SUM_TOTAL = 5
SUCCESS_COUNT = 0
FAILURE_SUM_TOTAL = 0
FAILURE_COUNT = 4
DATA_FRAME_WITH_NULLS = pd.DataFrame(
    {
        COL_A_WITH_NULLS: [None, 2, None, 4, None, None],
        COL_B_WITH_NULLS: [4, 3, None, None, None, None],
    }
)


class TestMultiColumnSumEqualsUnexpectedCount:
    @parameterize_batch_for_data_sources(
        data_source_configs=PANDAS_DATA_SOURCES + SPARK_DATA_SOURCES + SQL_DATA_SOURCES,
        data=DATA_FRAME,
    )
    def test_success(self, batch_for_datasource: Batch) -> None:
        batch = batch_for_datasource
        metric = MultiColumnSumEqualUnexpectedCount(
            sum_total=SUCCESS_SUM_TOTAL,
            column_list=[COL_A, COL_B],
        )
        result = batch.compute_metrics(metric)
        assert isinstance(result, MultiColumnSumEqualUnexpectedCountResult)
        assert result.value == SUCCESS_COUNT

    @parameterize_batch_for_data_sources(
        data_source_configs=PANDAS_DATA_SOURCES + SPARK_DATA_SOURCES + SQL_DATA_SOURCES,
        data=DATA_FRAME,
    )
    def test_failure(self, batch_for_datasource: Batch) -> None:
        batch = batch_for_datasource
        metric = MultiColumnSumEqualUnexpectedCount(
            sum_total=FAILURE_SUM_TOTAL,
            column_list=[COL_A, COL_B],
        )
        result = batch.compute_metrics(metric)
        assert isinstance(result, MultiColumnSumEqualUnexpectedCountResult)
        assert result.value == FAILURE_COUNT

    @pytest.mark.parametrize(
        "ignore_row_if,unexpected_count",
        [
            ("any_value_is_missing", 1),
            ("all_values_are_missing", 3),
            ("never", 6),
        ],
    )
    @parameterize_batch_for_data_sources(
        data_source_configs=PANDAS_DATA_SOURCES,
        data=DATA_FRAME_WITH_NULLS,
    )
    def test_ignore_row_if__pandas(
        self, batch_for_datasource: Batch, ignore_row_if, unexpected_count
    ) -> None:
        batch = batch_for_datasource
        metric = MultiColumnSumEqualUnexpectedCount(
            sum_total=FAILURE_SUM_TOTAL,
            column_list=[COL_A_WITH_NULLS, COL_B_WITH_NULLS],
            ignore_row_if=ignore_row_if,
        )
        result = batch.compute_metrics(metric)
        assert result.value == unexpected_count

    @pytest.mark.parametrize(
        "ignore_row_if,unexpected_count",
        [
            pytest.param(
                "any_value_is_missing", 1, marks=pytest.mark.xfail(reason="returns 6", strict=True)
            ),
            pytest.param(
                "all_values_are_missing",
                3,
                marks=pytest.mark.xfail(reason="returns 6", strict=True),
            ),
            pytest.param("never", 6),
        ],
    )
    @parameterize_batch_for_data_sources(
        data_source_configs=SPARK_DATA_SOURCES,
        data=DATA_FRAME_WITH_NULLS,
    )
    def test_ignore_row_if__spark(
        self, batch_for_datasource: Batch, ignore_row_if, unexpected_count
    ) -> None:
        batch = batch_for_datasource
        metric = MultiColumnSumEqualUnexpectedCount(
            sum_total=FAILURE_SUM_TOTAL,
            column_list=[COL_A_WITH_NULLS, COL_B_WITH_NULLS],
            ignore_row_if=ignore_row_if,
        )
        result = batch.compute_metrics(metric)
        assert result.value == unexpected_count

    @pytest.mark.parametrize(
        "ignore_row_if,unexpected_count",
        [
            pytest.param("any_value_is_missing", 1),
            pytest.param(
                "all_values_are_missing",
                3,
                marks=pytest.mark.xfail(reason="returns 1", strict=True),
            ),
            pytest.param("never", 6, marks=pytest.mark.xfail(reason="returns 1", strict=True)),
        ],
    )
    @parameterize_batch_for_data_sources(
        data_source_configs=SQL_DATA_SOURCES,
        data=DATA_FRAME_WITH_NULLS,
    )
    def test_ignore_row_if__sql(
        self, batch_for_datasource: Batch, ignore_row_if, unexpected_count
    ) -> None:
        batch = batch_for_datasource
        metric = MultiColumnSumEqualUnexpectedCount(
            sum_total=FAILURE_SUM_TOTAL,
            column_list=[COL_A_WITH_NULLS, COL_B_WITH_NULLS],
            ignore_row_if=ignore_row_if,
        )
        result = batch.compute_metrics(metric)
        assert result.value == unexpected_count



class MyCustomNonCoreGeClass:
    """
    This class is used only for testing.
    E.g. ensuring appropriate usage stats messaging when using plugin functionality when the custom class is not a core GX class type.
    """  # noqa: E501 # FIXME CoP



from __future__ import annotations

from great_expectations.compatibility.sqlalchemy import sqlalchemy as sa
from great_expectations.execution_engine import (
    PandasExecutionEngine,
    SparkDFExecutionEngine,
    SqlAlchemyExecutionEngine,
)
from great_expectations.expectations.metrics.map_metric_provider import (
    ColumnPairMapMetricProvider,
    column_pair_condition_partial,
)


class ColumnPairValuesAGreaterThanB(ColumnPairMapMetricProvider):
    condition_metric_name = "column_pair_values.a_greater_than_b"
    condition_domain_keys = (
        "batch_id",
        "table",
        "column_A",
        "column_B",
        "row_condition",
        "condition_parser",
        "ignore_row_if",
    )
    condition_value_keys = ("or_equal",)

    # noinspection PyPep8Naming
    @column_pair_condition_partial(engine=PandasExecutionEngine)
    def _pandas(cls, column_A, column_B, **kwargs):
        temp_column_A = column_A
        temp_column_B = column_B

        or_equal: bool = kwargs.get("or_equal") or False
        if or_equal:
            return temp_column_A >= temp_column_B
        else:
            return temp_column_A > temp_column_B

    # noinspection PyPep8Naming
    @column_pair_condition_partial(engine=SqlAlchemyExecutionEngine)
    def _sqlalchemy(cls, column_A, column_B, **kwargs):
        or_equal: bool = kwargs.get("or_equal") or False
        if or_equal:
            return sa.or_(
                column_A >= column_B,
                sa.and_(column_A == None, column_B == None),  # noqa: E711 # FIXME CoP
            )
        else:
            return column_A > column_B

    # noinspection PyPep8Naming
    @column_pair_condition_partial(engine=SparkDFExecutionEngine)
    def _spark(cls, column_A, column_B, **kwargs):
        temp_column_A = column_A
        temp_column_B = column_B

        or_equal: bool = kwargs.get("or_equal") or False
        if or_equal:
            return (temp_column_A >= temp_column_B) | (temp_column_A.eqNullSafe(temp_column_B))
        else:
            return temp_column_A > temp_column_B



import pathlib
from dataclasses import dataclass, field
from typing import Any, Mapping, Optional

import pandas as pd
import pytest

from great_expectations.compatibility.typing_extensions import override
from great_expectations.data_context import AbstractDataContext
from great_expectations.datasource.fluent.data_asset.path.pandas.generated_assets import CSVAsset
from great_expectations.datasource.fluent.interfaces import Batch
from tests.integration.sql_session_manager import SessionSQLEngineManager
from tests.integration.test_utils.data_source_config.base import (
    BatchTestSetup,
    DataSourceTestConfig,
)


@dataclass(frozen=True)
class PandasFilesystemCsvDatasourceTestConfig(DataSourceTestConfig):
    # see options: https://pandas.pydata.org/docs/reference/api/pandas.read_csv.html
    read_options: dict[str, Any] = field(default_factory=dict)
    # see options: https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.to_csv.html
    write_options: dict[str, Any] = field(default_factory=dict)

    @property
    @override
    def label(self) -> str:
        return "pandas-filesystem-csv"

    @property
    @override
    def pytest_mark(self) -> pytest.MarkDecorator:
        return pytest.mark.filesystem

    @override
    def create_batch_setup(
        self,
        request: pytest.FixtureRequest,
        data: pd.DataFrame,
        extra_data: Mapping[str, pd.DataFrame],
        context: AbstractDataContext,
        engine_manager: Optional[SessionSQLEngineManager] = None,
    ) -> BatchTestSetup:
        assert not extra_data, "extra_data is not supported for this data source."
        tmp_path = request.getfixturevalue("tmp_path")
        assert isinstance(tmp_path, pathlib.Path)

        return PandasFilesystemCsvBatchTestSetup(
            data=data,
            config=self,
            base_dir=tmp_path,
            context=context,
        )


class PandasFilesystemCsvBatchTestSetup(
    BatchTestSetup[PandasFilesystemCsvDatasourceTestConfig, CSVAsset]
):
    def __init__(
        self,
        config: PandasFilesystemCsvDatasourceTestConfig,
        data: pd.DataFrame,
        base_dir: pathlib.Path,
        context: AbstractDataContext,
    ) -> None:
        super().__init__(config=config, data=data, context=context)
        self._base_dir = base_dir

    @override
    def make_asset(self) -> CSVAsset:
        return self.context.data_sources.add_pandas_filesystem(
            name=self._random_resource_name(), base_directory=self._base_dir
        ).add_csv_asset(
            name=self._random_resource_name(),
            **self.config.read_options,
        )

    @override
    def make_batch(self) -> Batch:
        return (
            self.make_asset()
            .add_batch_definition_path(name=self._random_resource_name(), path=self.csv_path)
            .get_batch()
        )

    @override
    def setup(self) -> None:
        file_path = self._base_dir / self.csv_path
        self.data.to_csv(file_path, index=False, **self.config.write_options)

    @override
    def teardown(self) -> None: ...

    @property
    def csv_path(self) -> pathlib.Path:
        return pathlib.Path("data.csv")



from __future__ import annotations

import inspect
import logging
from dataclasses import dataclass, field
from enum import Enum
from typing import List, Optional, Sequence, Union

from typing_extensions import TypedDict

from great_expectations.core.expectation_validation_result import (
    ExpectationValidationResult,  # noqa: TC001 # FIXME CoP
)
from great_expectations.types import SerializableDictDot


class Maturity(str, Enum):
    """The four levels of maturity for features within Great Expectations"""

    CONCEPT_ONLY = "CONCEPT_ONLY"
    EXPERIMENTAL = "EXPERIMENTAL"
    BETA = "BETA"
    PRODUCTION = "PRODUCTION"


@dataclass
class AugmentedLibraryMetadata(SerializableDictDot):
    """An augmented version of the Expectation.library_metadata object, used within ExpectationDiagnostics"""  # noqa: E501 # FIXME CoP

    maturity: Maturity
    tags: List[str]
    contributors: List[str]
    requirements: List[str]
    library_metadata_passed_checks: bool
    has_full_test_suite: bool
    manually_reviewed_code: bool
    problems: List[str] = field(default_factory=list)

    legacy_maturity_level_substitutions = {
        "experimental": "EXPERIMENTAL",
        "beta": "BETA",
        "production": "PRODUCTION",
    }

    @classmethod
    def from_legacy_dict(cls, dict):
        """This method is a temporary adapter to allow typing of legacy library_metadata objects, without needing to immediately clean up every object."""  # noqa: E501 # FIXME CoP
        temp_dict = {}
        for k, v in dict.items():
            # Ignore parameters that don't match the type definition
            if k in inspect.signature(cls).parameters:
                temp_dict[k] = v
            else:
                logging.warning(
                    f"WARNING: Got extra parameter: {k} while instantiating AugmentedLibraryMetadata."  # noqa: E501 # FIXME CoP
                    "This parameter will be ignored."
                    "You probably need to clean up a library_metadata object."
                )

            # If necessary, substitute strings for precise Enum values.
            if (
                "maturity" in temp_dict
                and temp_dict["maturity"] in cls.legacy_maturity_level_substitutions
            ):
                temp_dict["maturity"] = cls.legacy_maturity_level_substitutions[
                    temp_dict["maturity"]
                ]

        return cls(**temp_dict)


@dataclass
class ExpectationDescriptionDiagnostics(SerializableDictDot):
    """Captures basic descriptive info about an Expectation. Used within the ExpectationDiagnostic object."""  # noqa: E501 # FIXME CoP

    camel_name: str
    snake_name: str
    short_description: str
    docstring: str


@dataclass
class RendererTestDiagnostics(SerializableDictDot):
    """Captures information from executing Renderer test cases. Used within the ExpectationRendererDiagnostics object."""  # noqa: E501 # FIXME CoP

    test_title: str
    rendered_successfully: bool
    renderered_str: Union[str, None]
    error_message: Union[str, None] = None
    stack_trace: Union[str, None] = None


@dataclass
class ExpectationRendererDiagnostics(SerializableDictDot):
    """Captures information about a specific Renderer within an Expectation. Used within the ExpectationDiagnostic object."""  # noqa: E501 # FIXME CoP

    name: str
    is_supported: bool
    is_standard: bool
    samples: List[RendererTestDiagnostics]


@dataclass
class ExpectationMetricDiagnostics(SerializableDictDot):
    """Captures information about a specific Metric dependency for an Expectation. Used within the ExpectationDiagnostic object."""  # noqa: E501 # FIXME CoP

    name: str
    has_question_renderer: bool


@dataclass
class ExpectationExecutionEngineDiagnostics(SerializableDictDot):
    """Captures which of the three Execution Engines are supported by an Expectation. Used within the ExpectationDiagnostic object."""  # noqa: E501 # FIXME CoP

    PandasExecutionEngine: bool
    SqlAlchemyExecutionEngine: bool
    SparkDFExecutionEngine: bool


@dataclass
class ExpectationErrorDiagnostics(SerializableDictDot):
    error_msg: str
    stack_trace: str
    test_title: Optional[str] = None
    test_backend: Optional[str] = None


@dataclass
class ExpectationTestDiagnostics(SerializableDictDot):
    """Captures information from executing Expectation test cases. Used within the ExpectationDiagnostic object."""  # noqa: E501 # FIXME CoP

    test_title: str
    backend: str
    test_passed: bool
    include_in_gallery: bool
    validation_result: Optional[ExpectationValidationResult]
    error_diagnostics: Optional[ExpectationErrorDiagnostics]


@dataclass
class ExpectationBackendTestResultCounts(SerializableDictDot):
    """Has each tested backend and the number of passing/failing tests"""

    backend: str
    num_passed: int
    num_failed: int
    failing_names: Optional[List[str]]


class ExpectationDiagnosticCheckMessageDict(TypedDict):
    message: str
    passed: bool


@dataclass
class ExpectationDiagnosticCheckMessage(SerializableDictDot):
    """Summarizes the result of a diagnostic Check. Used within the ExpectationDiagnostic object."""

    message: str
    passed: bool
    doc_url: Optional[str] = None
    sub_messages: Sequence[
        ExpectationDiagnosticCheckMessage | ExpectationDiagnosticCheckMessageDict
    ] = field(default_factory=list)


@dataclass
class ExpectationDiagnosticMaturityMessages(SerializableDictDot):
    """A holder for ExpectationDiagnosticCheckMessages, grouping them by maturity level. Used within the ExpectationDiagnostic object."""  # noqa: E501 # FIXME CoP

    experimental: List[ExpectationDiagnosticCheckMessage]
    beta: List[ExpectationDiagnosticCheckMessage]
    production: List[ExpectationDiagnosticCheckMessage]



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Iterable

from great_expectations._docs_decorators import public_api
from great_expectations.analytics.client import submit as submit_event
from great_expectations.analytics.events import (
    ExpectationSuiteCreatedEvent,
    ExpectationSuiteDeletedEvent,
)
from great_expectations.compatibility.pydantic import ValidationError as PydanticValidationError
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core import ExpectationSuite
from great_expectations.core.factory.factory import Factory
from great_expectations.data_context.data_context.context_factory import project_manager
from great_expectations.exceptions import DataContextError

if TYPE_CHECKING:
    from great_expectations.data_context.store import ExpectationsStore


@public_api
class SuiteFactory(Factory[ExpectationSuite]):
    """
    Responsible for basic CRUD operations on a Data Context's ExpectationSuites.
    """

    def __init__(self, store: ExpectationsStore):
        self._store = store

    @property
    def _include_rendered_content(self) -> bool:
        return project_manager.is_using_cloud()

    @public_api
    @override
    def add(self, suite: ExpectationSuite) -> ExpectationSuite:
        """Add an ExpectationSuite to the collection.

        Args:
            suite: ExpectationSuite to add

        Raises:
            DataContextError: if ExpectationSuite already exists
        """
        key = self._store.get_key(name=suite.name, id=None)
        if self._store.has_key(key=key):
            raise DataContextError(  # noqa: TRY003 # FIXME CoP
                f"Cannot add ExpectationSuite with name {suite.name} because it already exists."
            )
        self._store.add(key=key, value=suite)

        submit_event(
            event=ExpectationSuiteCreatedEvent(
                expectation_suite_id=suite.id,
            )
        )

        if suite._include_rendered_content:
            suite.render()

        return suite

    @public_api
    @override
    def delete(self, name: str) -> None:
        """Delete an ExpectationSuite from the collection.

        Args:
            name: The name of the ExpectationSuite to delete

        Raises:
            DataContextError: if ExpectationSuite doesn't exist
        """
        try:
            suite = self.get(name=name)
        except DataContextError as e:
            raise DataContextError(  # noqa: TRY003 # FIXME CoP
                f"Cannot delete ExpectationSuite with name {name} because it cannot be found."
            ) from e

        key = self._store.get_key(name=suite.name, id=suite.id)
        self._store.remove_key(key=key)

        submit_event(
            event=ExpectationSuiteDeletedEvent(
                expectation_suite_id=suite.id,
            )
        )

    @public_api
    @override
    def get(self, name: str) -> ExpectationSuite:
        """Get an ExpectationSuite from the collection by name.

        Args:
            name: Name of ExpectationSuite to get

        Raises:
            DataContextError: when ExpectationSuite is not found.
        """

        key = self._store.get_key(name=name, id=None)
        if not self._store.has_key(key=key):
            raise DataContextError(f"ExpectationSuite with name {name} was not found.")  # noqa: TRY003 # FIXME CoP
        suite_dict = self._store.get(key=key)
        return self._store.deserialize_suite_dict(suite_dict)

    @public_api
    @override
    def all(self) -> Iterable[ExpectationSuite]:
        """Get all ExpectationSuites."""
        dicts = self._store.get_all()
        # Marshmallow validation was done in the previous get_all() call for
        # suites but we can still die here because pydantic validation happens
        # on the expectations inside the suites here.
        # TODO: deserialization should not live in the factory and should
        # TODO: live in the store like in other domain objects. That will
        # TODO: allow us delete this error handling here.
        deserializable_suites: list[ExpectationSuite] = []
        bad_dicts: list[Any] = []
        for suite_dict in dicts:
            try:
                deserializable_suites.append(self._store.deserialize_suite_dict(suite_dict))
            except PydanticValidationError as e:
                bad_dicts.append(suite_dict)
                self._store.submit_all_deserialization_event(e)
            except Exception as e:
                self._store.submit_all_deserialization_event(e)
                raise
        return deserializable_suites

    @public_api
    @override
    def add_or_update(self, suite: ExpectationSuite) -> ExpectationSuite:
        """Add or update an ExpectationSuite by name.

        If an ExpectationSuite with the same name exists, overwrite it, otherwise
        create a new ExpectationSuite. On update, Expectations in the Suite which
        match a previously existing Expectation maintain a stable ID, and
        Expectations which have changed receive a new ID.

        Args:
            suite: ExpectationSuite to add or update
        """
        try:
            existing_suite = self.get(name=suite.name)
        except DataContextError:
            return self.add(suite=suite)

        # add IDs to expectations that haven't changed
        existing_expectations = existing_suite.expectations
        for expectation in suite.expectations:
            try:
                index = existing_expectations.index(expectation)
                expectation.id = existing_expectations[index].id
            except ValueError:
                pass  # expectation is new or updated

        suite.id = existing_suite.id
        suite.save()
        return suite



from __future__ import annotations

import itertools
import logging
from typing import Any, Dict, List, Optional, Sequence, Union

import pytest

import great_expectations.expectations as gxe
from great_expectations.compatibility import pydantic
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.expectation_validation_result import ExpectationValidationResult
from great_expectations.exceptions import InvalidExpectationConfigurationError
from great_expectations.execution_engine.execution_engine import ExecutionEngine
from great_expectations.expectations.expectation import (
    ColumnMapExpectation,
    ColumnPairMapExpectation,
    Expectation,
    MulticolumnMapExpectation,
    _validate_dependencies_against_available_metrics,
)
from great_expectations.expectations.expectation_configuration import (
    ExpectationConfiguration,
)
from great_expectations.expectations.model_field_types import (
    MostlyField,  # type needed in pydantic validation
    ValueSetField,  # type needed in pydantic validation
)
from great_expectations.expectations.window import Offset, Window
from great_expectations.validator.metric_configuration import MetricConfiguration

LOGGER = logging.getLogger(__name__)


class FakeMulticolumnExpectation(MulticolumnMapExpectation):
    map_metric = "fake_multicol_metric"


class FakeColumnMapExpectation(ColumnMapExpectation):
    map_metric = "fake_col_metric"


class FakeColumnPairMapExpectation(ColumnPairMapExpectation):
    map_metric = "fake_pair_metric"


@pytest.fixture
def metrics_dict():
    """
    Fixture for metrics dict, which represents Metrics already calculated for given Batch
    """
    return {
        (
            "column_values.nonnull.unexpected_count",
            "e197e9d84e4f8aa077b8dd5f9042b382",
            (),
        ): "i_exist"
    }


def fake_metrics_config_list(
    metric_name: str, metric_domain_kwargs: Dict[str, Any]
) -> List[MetricConfiguration]:
    """
    Helper method to generate list of MetricConfiguration objects for tests.
    """
    return [
        MetricConfiguration(
            metric_name=metric_name,
            metric_domain_kwargs=metric_domain_kwargs,
            metric_value_kwargs={},
        )
    ]


def fake_expectation_config(
    expectation_type: str, config_kwargs: Dict[str, Any]
) -> ExpectationConfiguration:
    """
    Helper method to generate of ExpectationConfiguration objects for tests.
    """
    return ExpectationConfiguration(
        type=expectation_type,
        kwargs=config_kwargs,
    )


@pytest.mark.unit
@pytest.mark.parametrize(
    "fake_expectation_cls, config",
    [
        (
            FakeMulticolumnExpectation,
            fake_expectation_config(
                "fake_multicolumn_expectation", {"column_list": ["column_1", "column_2"]}
            ),
        ),
        (
            FakeColumnMapExpectation,
            fake_expectation_config("fake_column_map_expectation", {"column": "col"}),
        ),
        (
            FakeColumnPairMapExpectation,
            fake_expectation_config(
                "fake_column_pair_map_expectation",
                {"column_A": "colA", "column_B": "colB"},
            ),
        ),
    ],
)
def test_multicolumn_expectation_has_default_mostly(fake_expectation_cls, config):
    try:
        fake_expectation = fake_expectation_cls(**config.kwargs)
    except Exception:
        assert False, "Validate configuration threw an error when testing default mostly value"
    assert fake_expectation._get_success_kwargs().get("mostly") == 1, (
        "Default mostly success ratio is not 1"
    )


@pytest.mark.unit
@pytest.mark.parametrize(
    "fake_expectation_cls, config",
    itertools.chain(
        *[
            [
                (
                    FakeMulticolumnExpectation,
                    fake_expectation_config(
                        "fake_multicolumn_expectation",
                        {"column_list": ["column_1", "column_2"], "mostly": x},
                    ),
                )
                for x in [0, 0.5, 1]
            ],
            [
                (
                    FakeColumnMapExpectation,
                    fake_expectation_config(
                        "fake_column_map_expectation", {"column": "col", "mostly": x}
                    ),
                )
                for x in [0, 0.5, 1]
            ],
            [
                (
                    FakeColumnPairMapExpectation,
                    fake_expectation_config(
                        "fake_column_pair_map_expectation",
                        {"column_A": "colA", "column_B": "colB", "mostly": x},
                    ),
                )
                for x in [0, 0.5, 1]
            ],
        ]
    ),
)
def test_expectation_succeeds_with_valid_mostly(fake_expectation_cls, config):
    fake_expectation = fake_expectation_cls(**config.kwargs)
    assert fake_expectation._get_success_kwargs().get("mostly") == config.kwargs["mostly"], (
        "Default mostly success ratio is not 1"
    )


@pytest.mark.unit
@pytest.mark.parametrize(
    "fake_expectation_cls, config",
    [
        (
            FakeMulticolumnExpectation,
            fake_expectation_config(
                "fake_multicolumn_expectation",
                {"column_list": ["column_1", "column_2"], "mostly": -0.5},
            ),
        ),
        (
            FakeColumnMapExpectation,
            fake_expectation_config(
                "fake_column_map_expectation", {"column": "col", "mostly": 1.5}
            ),
        ),
        (
            FakeColumnPairMapExpectation,
            fake_expectation_config(
                "fake_column_pair_map_expectation",
                {"column_A": "colA", "column_B": "colB", "mostly": -1},
            ),
        ),
    ],
)
def test_multicolumn_expectation_validation_errors_with_bad_mostly(fake_expectation_cls, config):
    with pytest.raises(pydantic.ValidationError):
        fake_expectation_cls(**config)


@pytest.mark.unit
def test_validate_dependencies_against_available_metrics_success(metrics_dict):
    metric_config_list: List[MetricConfiguration] = fake_metrics_config_list(
        metric_name="column_values.nonnull.unexpected_count",
        metric_domain_kwargs={
            "batch_id": "projects-projects",
            "column": "i_exist",
        },
    )
    _validate_dependencies_against_available_metrics(
        validation_dependencies=metric_config_list,
        metrics=metrics_dict,
    )


@pytest.mark.unit
def test_validate_dependencies_against_available_metrics_failure(metrics_dict):
    metric_config_list: List[MetricConfiguration] = fake_metrics_config_list(
        metric_name="column_values.nonnull.unexpected_count",
        metric_domain_kwargs={
            "batch_id": "projects-projects",
            "column": "i_dont_exist",
        },
    )
    with pytest.raises(InvalidExpectationConfigurationError):
        _validate_dependencies_against_available_metrics(
            validation_dependencies=metric_config_list,
            metrics=metrics_dict,
        )


@pytest.mark.unit
def test_expectation_configuration_property():
    expectation = gxe.ExpectColumnMaxToBeBetween(column="foo", min_value=0, max_value=10)

    assert expectation.configuration == ExpectationConfiguration(
        type="expect_column_max_to_be_between",
        kwargs={
            "column": "foo",
            "min_value": 0,
            "max_value": 10,
        },
    )


@pytest.mark.unit
def test_expectation_configuration_window():
    expectation = gxe.ExpectColumnMaxToBeBetween(
        column="foo",
        min_value=0,
        max_value=10,
        windows=[
            Window(
                constraint_fn="a",
                parameter_name="b",
                range=5,
                offset=Offset(positive=0.2, negative=0.2),
                strict=True,
            )
        ],
    )

    assert expectation.configuration == ExpectationConfiguration(
        type="expect_column_max_to_be_between",
        kwargs={
            "column": "foo",
            "min_value": 0,
            "max_value": 10,
            "windows": [
                {
                    "constraint_fn": "a",
                    "parameter_name": "b",
                    "range": 5,
                    "offset": {"positive": 0.2, "negative": 0.2},
                    "strict": True,
                }
            ],
        },
    )


@pytest.mark.unit
def test_expectation_configuration_window_empty():
    expectation = gxe.ExpectColumnMaxToBeBetween(
        column="foo",
        min_value=0,
        max_value=10,
        windows=None,
    )

    assert expectation.configuration == ExpectationConfiguration(
        type="expect_column_max_to_be_between",
        kwargs={
            "column": "foo",
            "min_value": 0,
            "max_value": 10,
        },
    )


@pytest.mark.unit
def test_expectation_configuration_property_recognizes_state_changes():
    expectation = gxe.ExpectColumnMaxToBeBetween(column="foo", min_value=0, max_value=10)

    expectation.column = "bar"
    expectation.min_value = 5
    expectation.max_value = 15

    assert expectation.configuration == ExpectationConfiguration(
        type="expect_column_max_to_be_between",
        kwargs={
            "column": "bar",
            "min_value": 5,
            "max_value": 15,
        },
    )


@pytest.mark.unit
def test_unrecognized_expectation_arg_raises_error():
    with pytest.raises(pydantic.ValidationError, match="extra fields not permitted"):
        gxe.ExpectColumnMaxToBeBetween(
            column="foo",
            min_value=0,
            max_value=10,
            mostyl=0.95,  # 'mostly' typo
        )


class TestSuiteParameterOptions:
    """Tests around the suite_parameter_options property of Expectations.

    Note: evaluation_parameter_options is currently a sorted tuple, but doesn't necessarily have to be
    """  # noqa: E501 # FIXME CoP

    SUITE_PARAMETER_MIN = "my_min"
    SUITE_PARAMETER_MAX = "my_max"
    SUITE_PARAMETER_VALUE = "my_value"
    SUITE_PARAMETER_MOSTLY = "my_mostly"

    @pytest.mark.unit
    def test_expectation_without_evaluation_parameter(self):
        expectation = gxe.ExpectColumnValuesToBeBetween(column="foo", min_value=0, max_value=10)
        assert expectation.suite_parameter_options == tuple()

    @pytest.mark.unit
    def test_expectation_with_evaluation_parameter(self):
        expectation = gxe.ExpectColumnValuesToBeBetween(
            column="foo",
            min_value=0,
            max_value={"$PARAMETER": self.SUITE_PARAMETER_MAX},
        )
        assert expectation.suite_parameter_options == (self.SUITE_PARAMETER_MAX,)

    @pytest.mark.unit
    def test_column_map_expectation_with_evaluation_parameter(self):
        expectation = gxe.ExpectColumnValuesToBeNull(
            column="foo", mostly={"$PARAMETER": self.SUITE_PARAMETER_MOSTLY}
        )
        assert expectation.suite_parameter_options == (self.SUITE_PARAMETER_MOSTLY,)

    @pytest.mark.unit
    def test_expectation_with_multiple_suite_parameters(self):
        expectation = gxe.ExpectColumnValuesToBeBetween(
            column="foo",
            min_value={"$PARAMETER": self.SUITE_PARAMETER_MIN},
            max_value={"$PARAMETER": self.SUITE_PARAMETER_MAX},
        )
        assert expectation.suite_parameter_options == (
            self.SUITE_PARAMETER_MAX,
            self.SUITE_PARAMETER_MIN,
        )

    @pytest.mark.unit
    def test_expectation_with_duplicate_suite_parameters(self):
        expectation = gxe.ExpectColumnValuesToBeBetween(
            column="foo",
            min_value={"$PARAMETER": self.SUITE_PARAMETER_VALUE},
            max_value={"$PARAMETER": self.SUITE_PARAMETER_VALUE},
        )
        assert expectation.suite_parameter_options == (self.SUITE_PARAMETER_VALUE,)


@pytest.mark.unit
@pytest.mark.parametrize(
    "column_a,column_b,expected",
    [
        pytest.param("foo", "foo", True, id="equivalent_columns"),
        pytest.param("foo", "bar", False, id="different_columns"),
    ],
)
def test_expectation_equality(column_a: str, column_b: str, expected: bool):
    expectation_a = gxe.ExpectColumnValuesToBeBetween(column=column_a, min_value=0, max_value=10)
    expectation_b = gxe.ExpectColumnValuesToBeBetween(column=column_b, min_value=0, max_value=10)

    assert (expectation_a == expectation_b) is expected


@pytest.mark.unit
@pytest.mark.parametrize(
    "notes_a,notes_b,expected",
    [
        pytest.param(None, None, True, id="both_none"),
        pytest.param([], None, True, id="both_falsy"),
        pytest.param("my_notes", None, False, id="missing_notes"),
        pytest.param("my_notes", "my_other_notes", False, id="different_notes"),
        pytest.param("my_notes", "my_notes", True, id="equivalent_notes"),
    ],
)
def test_expectation_equality_with_notes(
    notes_a: str | list[str] | None, notes_b: str | list[str] | None, expected: bool
):
    expectation_a = gxe.ExpectColumnValuesToBeBetween(
        column="foo", min_value=0, max_value=10, notes=notes_a
    )
    expectation_b = gxe.ExpectColumnValuesToBeBetween(
        column="foo", min_value=0, max_value=10, notes=notes_b
    )

    assert (expectation_a == expectation_b) is expected


@pytest.mark.unit
@pytest.mark.parametrize(
    "meta_a,meta_b,expected",
    [
        pytest.param(None, None, True, id="both_none"),
        pytest.param({}, None, True, id="both_falsy"),
        pytest.param({"author": "Bob Dylan"}, None, False, id="missing_meta"),
        pytest.param(
            {"author": "Bob Dylan"}, {"author": "John Lennon"}, False, id="different_meta"
        ),
        pytest.param({"author": "Bob Dylan"}, {"author": "Bob Dylan"}, True, id="equivalent_meta"),
    ],
)
def test_expectation_equality_with_meta(meta_a: dict | None, meta_b: dict | None, expected: bool):
    expectation_a = gxe.ExpectColumnValuesToBeBetween(
        column="foo", min_value=0, max_value=10, meta=meta_a
    )
    expectation_b = gxe.ExpectColumnValuesToBeBetween(
        column="foo", min_value=0, max_value=10, meta=meta_b
    )

    assert (expectation_a == expectation_b) is expected


@pytest.mark.unit
def test_expectation_equality_ignores_rendered_content():
    column = "whatever"
    min_value = 0
    max_value = 10
    expectation_a = gxe.ExpectColumnValuesToBeBetween(
        column=column,
        min_value=min_value,
        max_value=max_value,
    )
    expectation_a.render()
    assert expectation_a.rendered_content

    expectation_b = gxe.ExpectColumnValuesToBeBetween(
        column=column,
        min_value=min_value,
        max_value=max_value,
    )
    expectation_b.rendered_content = None

    assert expectation_a == expectation_b


@pytest.mark.unit
@pytest.mark.parametrize(
    "expectation_a, expectation_b, expected_result",
    [
        pytest.param(
            gxe.ExpectColumnValuesToBeBetween(column="foo", min_value=0),
            {},
            False,
            id="different_objects",
        ),
        pytest.param(
            gxe.ExpectColumnDistinctValuesToBeInSet(column="bar", value_set=[1, 2, 3]),
            gxe.ExpectColumnValuesToBeBetween(column="foo", min_value=0),
            True,
            id="different_expectation_types",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeBetween(column="foo", min_value=0),
            gxe.ExpectColumnValuesToBeBetween(column="foo", min_value=0),
            False,
            id="equivalent_expectations",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeBetween(
                column="foo", min_value=0, id="bbbe648e-0a43-431b-81a0-04e68f1473ae"
            ),
            gxe.ExpectColumnValuesToBeBetween(
                column="foo", min_value=0, id="aaae648e-0a43-431b-81a0-04e68f1473ae"
            ),
            False,
            id="equiv_expectations_with_ids",
        ),
    ],
)
def test_expectations___lt__(expectation_a, expectation_b, expected_result):
    assert (expectation_a < expectation_b) is expected_result


@pytest.mark.unit
def test_expectation_sorting():
    expectation_a = gxe.ExpectColumnValuesToBeBetween(
        column="foo", min_value=0, id="80b6d508-a843-426e-97c0-7ff64d35ac04"
    )
    expectation_b = gxe.ExpectColumnValuesToBeBetween(
        column="foo", min_value=0, id="4cd1e63a-880b-46ea-93e8-c11636df18b8"
    )
    expectation_c = gxe.ExpectTableColumnCountToBeBetween()
    expectation_d = gxe.ExpectColumnMaxToBeBetween(column="foo", min_value=0, max_value=10)
    expectation_e = gxe.ExpectColumnMedianToBeBetween(column="foo", min_value=0, max_value=10)

    expectations = [expectation_a, expectation_b, expectation_c, expectation_d, expectation_e]

    assert sorted(expectations) == [
        expectation_d,
        expectation_e,
        expectation_b,
        expectation_a,
        expectation_c,
    ]


class _SampleExpectation(Expectation):
    mostly: MostlyField
    value_set: ValueSetField

    @override
    def _validate(
        self,
        metrics: dict,
        runtime_configuration: Optional[dict] = None,
        execution_engine: Optional[ExecutionEngine] = None,
    ) -> Union[ExpectationValidationResult, dict]:
        # just satisfying the abc
        return {}


class TestCustomAnnotatedFields:
    @pytest.mark.parametrize(
        "mostly",
        [
            0,
            0.0,
            0.5,
            0.55555,
            1,
            1.0,
        ],
    )
    @pytest.mark.unit
    def test_valid_mostly_values(self, mostly: float):
        expectation = _SampleExpectation(mostly=mostly, value_set=[1, 2, 3])
        assert expectation.mostly == mostly

    @pytest.mark.parametrize(
        "mostly",
        [
            None,
            "one",
            -1,
            2,
        ],
    )
    @pytest.mark.unit
    def test_invalid_mostly_values(self, mostly: Any):
        with pytest.raises(pydantic.ValidationError):
            _SampleExpectation(mostly=mostly, value_set=[1, 2, 3])

    @pytest.mark.parametrize(
        "value_set,expected_value",
        [
            (["a"], ["a"]),
            ([1], [1]),
            ({"a"}, ["a"]),
            ({1}, [1]),
            ([1, 2, 3], [1, 2, 3]),
            (["a", "b", "c"], ["a", "b", "c"]),
            ({"$PARAMETER": "my_param"}, {"$PARAMETER": "my_param"}),
        ],
    )
    @pytest.mark.unit
    def test_valid_value_set_values(self, value_set: Union[Sequence, set], expected_value: Any):
        expectation = _SampleExpectation(mostly=1, value_set=value_set)
        assert expectation.value_set == expected_value

    @pytest.mark.parametrize(
        "input_value,expected_type",
        [
            ({"1", "2", "3"}, list),
            (("1", "2", "3"), list),
            (["1", "2", "3"], list),
            ({"$PARAMETER": "my_param"}, dict),
        ],
    )
    @pytest.mark.unit
    def test_value_set_field_converts_to_list(self, input_value: Any, expected_type: type) -> None:
        expectation = _SampleExpectation(mostly=1, value_set=input_value)
        assert isinstance(expectation.value_set, expected_type)


class TestExpectationHash:
    @pytest.mark.unit
    def test_hash_consistency_with_equality(self):
        expectation1 = gxe.ExpectColumnValuesToNotBeNull(
            column="test_column", mostly=0.95, meta={"test": "value"}, notes="test notes"
        )
        expectation2 = gxe.ExpectColumnValuesToNotBeNull(
            column="test_column", mostly=0.95, meta={"test": "value"}, notes="test notes"
        )

        assert expectation1 == expectation2
        assert hash(expectation1) == hash(expectation2)

    @pytest.mark.unit
    def test_hash_different_for_different_columns(self):
        expectation1 = gxe.ExpectColumnValuesToNotBeNull(column="test_column_1")
        expectation2 = gxe.ExpectColumnValuesToNotBeNull(column="test_column_2")

        assert expectation1 != expectation2
        assert hash(expectation1) != hash(expectation2)

    @pytest.mark.unit
    def test_hash_different_for_different_mostly(self):
        expectation1 = gxe.ExpectColumnValuesToNotBeNull(column="test_column", mostly=0.95)
        expectation2 = gxe.ExpectColumnValuesToNotBeNull(column="test_column", mostly=0.90)

        assert expectation1 != expectation2
        assert hash(expectation1) != hash(expectation2)

    @pytest.mark.unit
    def test_hash_different_for_different_meta(self):
        expectation1 = gxe.ExpectColumnValuesToNotBeNull(
            column="test_column", meta={"test": "value1"}
        )
        expectation2 = gxe.ExpectColumnValuesToNotBeNull(
            column="test_column", meta={"test": "value2"}
        )

        assert expectation1 != expectation2
        assert hash(expectation1) != hash(expectation2)

    @pytest.mark.unit
    def test_hash_excludes_rendered_content(self):
        expectation1 = gxe.ExpectColumnValuesToNotBeNull(column="test_column")
        expectation2 = gxe.ExpectColumnValuesToNotBeNull(column="test_column")

        expectation1.render()

        assert expectation1 == expectation2
        assert hash(expectation1) == hash(expectation2)

    @pytest.mark.unit
    def test_hash_stable_across_runs(self):
        expectation = gxe.ExpectColumnValuesToNotBeNull(
            column="test_column", mostly=0.95, meta={"test": "value"}, notes="test notes"
        )

        hash1 = hash(expectation)
        hash2 = hash(expectation)
        hash3 = hash(expectation)

        assert hash1 == hash2 == hash3



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    ClassVar,
    List,
    Type,
)

from great_expectations.datasource.fluent.data_asset.path.file_asset import (
    FileDataAsset,  # noqa: TC001  # pydantic requires this type at runtime
)
from great_expectations.datasource.fluent.data_asset.path.pandas.generated_assets import (
    _FILE_PATH_ASSET_MODELS,
)
from great_expectations.datasource.fluent.pandas_datasource import (
    _PandasDatasource,
)

if TYPE_CHECKING:
    from great_expectations.datasource.fluent.interfaces import DataAsset


class _PandasFilePathDatasource(_PandasDatasource):
    # class attributes
    asset_types: ClassVar[List[Type[DataAsset]]] = list(_FILE_PATH_ASSET_MODELS.values())

    # instance attributes
    assets: List[FileDataAsset] = []



from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar, Dict, Literal, Optional, Type, Union

from great_expectations.compatibility import pydantic
from great_expectations.core.suite_parameters import (
    SuiteParameterDict,  # noqa: TC001 # FIXME CoP
)
from great_expectations.expectations.expectation import (
    ColumnPairMapExpectation,
    render_suite_parameter_string,
)
from great_expectations.expectations.metadata_types import DataQualityIssues, SupportedDataSources
from great_expectations.expectations.model_field_descriptions import (
    COLUMN_A_DESCRIPTION,
    COLUMN_B_DESCRIPTION,
    IGNORE_ROW_IF_DESCRIPTION,
    MOSTLY_DESCRIPTION,
)
from great_expectations.render import LegacyRendererType, RenderedStringTemplateContent
from great_expectations.render.renderer.renderer import renderer
from great_expectations.render.renderer_configuration import (
    RendererConfiguration,
    RendererValueType,
)
from great_expectations.render.util import (
    num_to_str,
    parse_row_condition_string_pandas_engine,
    substitute_none_for_missing,
)

if TYPE_CHECKING:
    from great_expectations.core import (
        ExpectationValidationResult,
    )
    from great_expectations.expectations.expectation_configuration import (
        ExpectationConfiguration,
    )
    from great_expectations.render.renderer_configuration import AddParamArgs

EXPECTATION_SHORT_DESCRIPTION = "Expect the values in column A to be greater than column B."
OR_EQUAL_DESCRIPTION = "If True, then values can be equal, not strictly greater."
SUPPORTED_DATA_SOURCES = [
    SupportedDataSources.PANDAS.value,
    SupportedDataSources.SPARK.value,
    SupportedDataSources.POSTGRESQL.value,
    SupportedDataSources.MYSQL.value,
    SupportedDataSources.MSSQL.value,
    SupportedDataSources.BIGQUERY.value,
    SupportedDataSources.SNOWFLAKE.value,
    SupportedDataSources.DATABRICKS.value,
    SupportedDataSources.REDSHIFT.value,
]
DATA_QUALITY_ISSUES = [DataQualityIssues.NUMERIC.value]


class ExpectColumnPairValuesAToBeGreaterThanB(ColumnPairMapExpectation):
    __doc__ = f"""{EXPECTATION_SHORT_DESCRIPTION}

    ExpectColumnPairValuesAToBeGreaterThanB is a \
    Column Pair Map Expectation.

    Column Pair Map Expectations are evaluated for a pair of columns and ask a yes/no question about the row-wise relationship between those two columns.
    Based on the result, they then calculate the percentage of rows that gave a positive answer.
    If the percentage is high enough, the Expectation considers that data valid.

    Args:
        column_A (str): {COLUMN_A_DESCRIPTION}
        column_B (str): {COLUMN_B_DESCRIPTION}
        or_equal (boolean or None): {OR_EQUAL_DESCRIPTION}

    Other Parameters:
        ignore_row_if (str): \
            "both_values_are_missing", "either_value_is_missing", "neither" \
            {IGNORE_ROW_IF_DESCRIPTION} Default "neither".
        mostly (None or a float between 0 and 1): \
            {MOSTLY_DESCRIPTION} \
            For more detail, see [mostly](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#mostly). Default 1.
        result_format (str or None): \
            Which output mode to use: BOOLEAN_ONLY, BASIC, COMPLETE, or SUMMARY. \
            For more detail, see [result_format](https://docs.greatexpectations.io/docs/reference/expectations/result_format).
        catch_exceptions (boolean or None): \
            If True, then catch exceptions and include them as part of the result object. \
            For more detail, see [catch_exceptions](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#catch_exceptions).
        meta (dict or None): \
            A JSON-serializable dictionary (nesting allowed) that will be included in the output without modification. \
            For more detail, see [meta](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#meta).

    Returns:
        An [ExpectationSuiteValidationResult](https://docs.greatexpectations.io/docs/terms/validation_result)

        Exact fields vary depending on the values passed to result_format, catch_exceptions, and meta.

    Supported Data Sources:
        [{SUPPORTED_DATA_SOURCES[0]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[1]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[2]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[3]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[4]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[5]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[6]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[7]}](https://docs.greatexpectations.io/docs/application_integration_support/)


    Data Quality Issues:
        {DATA_QUALITY_ISSUES[0]}

    Example Data:
                test 	test2
            0 	2       1
            1 	2       2
            2 	4   	4

    Code Examples:
        Passing Case:
            Input:
                ExpectColumnPairValuesAToBeGreaterThanB(
                    column_A="test",
                    column_B="test2",
                    or_equal=True
            )

            Output:
                {{
                  "exception_info": {{
                    "raised_exception": false,
                    "exception_traceback": null,
                    "exception_message": null
                  }},
                  "result": {{
                    "element_count": 3,
                    "unexpected_count": 0,
                    "unexpected_percent": 0.0,
                    "partial_unexpected_list": [],
                    "missing_count": 0,
                    "missing_percent": 0.0,
                    "unexpected_percent_total": 0.0,
                    "unexpected_percent_nonmissing": 0.0
                  }},
                  "meta": {{}},
                  "success": true
                }}

        Failing Case:
            Input:
                ExpectColumnPairValuesAToBeGreaterThanB(
                    column_A="test2",
                    column_B="test"
            )

            Output:
                {{
                  "exception_info": {{
                    "raised_exception": false,
                    "exception_traceback": null,
                    "exception_message": null
                  }},
                  "result": {{
                    "element_count": 3,
                    "unexpected_count": 3,
                    "unexpected_percent": 100.0,
                    "partial_unexpected_list": [
                      [
                        1,
                        2
                      ],
                      [
                        2,
                        2
                      ],
                      [
                        4,
                        4
                      ]
                    ],
                    "missing_count": 0,
                    "missing_percent": 0.0,
                    "unexpected_percent_total": 100.0,
                    "unexpected_percent_nonmissing": 100.0
                  }},
                  "meta": {{}},
                  "success": false
                }}
    """  # noqa: E501 # FIXME CoP

    or_equal: Union[bool, SuiteParameterDict, None] = pydantic.Field(
        default=None, description=OR_EQUAL_DESCRIPTION
    )
    ignore_row_if: Union[
        Literal["both_values_are_missing", "either_value_is_missing", "neither"],
        SuiteParameterDict,
    ] = pydantic.Field(
        default="both_values_are_missing",
        description=IGNORE_ROW_IF_DESCRIPTION,
    )

    # This dictionary contains metadata for display in the public gallery
    library_metadata: ClassVar[Dict[str, Union[str, list, bool]]] = {
        "maturity": "production",
        "tags": [
            "core expectation",
            "column pair map expectation",
        ],
        "contributors": ["@great_expectations"],
        "requirements": [],
        "has_full_test_suite": True,
        "manually_reviewed_code": True,
    }
    _library_metadata = library_metadata

    map_metric = "column_pair_values.a_greater_than_b"
    success_keys = (
        "column_A",
        "column_B",
        "ignore_row_if",
        "or_equal",
        "mostly",
    )
    args_keys = (
        "column_A",
        "column_B",
        "or_equal",
    )

    class Config:
        title = "Expect column pair values A to be greater than B"

        @staticmethod
        def schema_extra(
            schema: Dict[str, Any], model: Type[ExpectColumnPairValuesAToBeGreaterThanB]
        ) -> None:
            ColumnPairMapExpectation.Config.schema_extra(schema, model)
            schema["properties"]["metadata"]["properties"].update(
                {
                    "data_quality_issues": {
                        "title": "Data Quality Issues",
                        "type": "array",
                        "const": DATA_QUALITY_ISSUES,
                    },
                    "library_metadata": {
                        "title": "Library Metadata",
                        "type": "object",
                        "const": model._library_metadata,
                    },
                    "short_description": {
                        "title": "Short Description",
                        "type": "string",
                        "const": EXPECTATION_SHORT_DESCRIPTION,
                    },
                    "supported_data_sources": {
                        "title": "Supported Data Sources",
                        "type": "array",
                        "const": SUPPORTED_DATA_SOURCES,
                    },
                }
            )

    @classmethod
    def _prescriptive_template(
        cls,
        renderer_configuration: RendererConfiguration,
    ) -> RendererConfiguration:
        add_param_args: AddParamArgs = (
            ("column_A", RendererValueType.STRING),
            ("column_B", RendererValueType.STRING),
            ("ignore_row_if", RendererValueType.STRING),
            ("mostly", RendererValueType.NUMBER),
            ("or_equal", RendererValueType.BOOLEAN),
        )
        for name, param_type in add_param_args:
            renderer_configuration.add_param(name=name, param_type=param_type)

        params = renderer_configuration.params
        template_str = ""

        if not params.column_A or not params.column_B:
            template_str += "$column has a bogus `expect_column_pair_values_A_to_be_greater_than_B` expectation. "  # noqa: E501 # FIXME CoP

        if not params.mostly or params.mostly.value == 1.0:
            if not params.or_equal:
                template_str += (
                    "Values in $column_A must always be greater than those in $column_B."
                )
            else:
                template_str += "Values in $column_A must always be greater than or equal to those in $column_B."  # noqa: E501 # FIXME CoP
        else:
            renderer_configuration = cls._add_mostly_pct_param(
                renderer_configuration=renderer_configuration
            )
            if not params.or_equal:
                template_str = "Values in $column_A must be greater than those in $column_B, at least $mostly_pct % of the time."  # noqa: E501 # FIXME CoP
            else:
                template_str = "Values in $column_A must be greater than or equal to those in $column_B, at least $mostly_pct % of the time."  # noqa: E501 # FIXME CoP

        renderer_configuration.template_str = template_str

        return renderer_configuration

    @classmethod
    @renderer(renderer_type=LegacyRendererType.PRESCRIPTIVE)
    @render_suite_parameter_string
    def _prescriptive_renderer(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
        **kwargs,
    ):
        runtime_configuration = runtime_configuration or {}
        _ = runtime_configuration.get("include_column_name") is not False
        styling = runtime_configuration.get("styling")
        params = substitute_none_for_missing(
            configuration.kwargs,
            [
                "column_A",
                "column_B",
                "ignore_row_if",
                "mostly",
                "or_equal",
                "row_condition",
                "condition_parser",
            ],
        )

        if (params["column_A"] is None) or (params["column_B"] is None):
            template_str = "$column has a bogus `expect_column_pair_values_A_to_be_greater_than_B` expectation."  # noqa: E501 # FIXME CoP
            params["row_condition"] = None

        if params["mostly"] is None or params["mostly"] == 1.0:
            if params["or_equal"] in [None, False]:
                template_str = "Values in $column_A must always be greater than those in $column_B."
            else:
                template_str = "Values in $column_A must always be greater than or equal to those in $column_B."  # noqa: E501 # FIXME CoP
        else:
            params["mostly_pct"] = num_to_str(params["mostly"] * 100, no_scientific=True)
            # params["mostly_pct"] = "{:.14f}".format(params["mostly"]*100).rstrip("0").rstrip(".")
            if params["or_equal"] in [None, False]:
                template_str = "Values in $column_A must be greater than those in $column_B, at least $mostly_pct % of the time."  # noqa: E501 # FIXME CoP
            else:
                template_str = "Values in $column_A must be greater than or equal to those in $column_B, at least $mostly_pct % of the time."  # noqa: E501 # FIXME CoP

        if params["row_condition"] is not None:
            (
                conditional_template_str,
                conditional_params,
            ) = parse_row_condition_string_pandas_engine(params["row_condition"])
            template_str = (
                conditional_template_str + ", then " + template_str[0].lower() + template_str[1:]
            )
            params.update(conditional_params)

        return [
            RenderedStringTemplateContent(
                **{
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": template_str,
                        "params": params,
                        "styling": styling,
                    },
                }
            )
        ]



from __future__ import annotations

import json

from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.id_dict import IDDict, IDDictID
from great_expectations.types.base import SerializableDotDict
from great_expectations.util import convert_to_json_serializable  # noqa: TID251 # FIXME CoP


class ExceptionInfo(SerializableDotDict):
    def __init__(
        self,
        exception_traceback: str,
        exception_message: str,
        raised_exception: bool = True,
    ) -> None:
        super().__init__(
            exception_traceback=exception_traceback,
            exception_message=exception_message,
            raised_exception=raised_exception,
        )

    @override
    def to_json_dict(self) -> dict:
        fields_dict: dict = {
            "exception_traceback": self.exception_traceback,
            "exception_message": self.exception_message,
            "raised_exception": self.raised_exception,
        }
        return convert_to_json_serializable(fields_dict)

    @property
    def exception_traceback(self) -> str:
        return self["exception_traceback"]

    @property
    def exception_message(self) -> str:
        return self["exception_message"]

    @property
    def raised_exception(self) -> bool:
        return self["raised_exception"]

    @override
    def __repr__(self) -> str:
        fields_dict: dict = {
            "exception_traceback": self.exception_traceback,
            "exception_message": self.exception_message,
            "raised_exception": self.raised_exception,
        }
        return str(fields_dict)

    @property
    def id(self) -> IDDictID:
        return IDDict(self.to_json_dict()).to_id()

    def __eq__(self, other):  # type: ignore[explicit-override] # FIXME
        if not isinstance(other, self.__class__):
            # Delegate comparison to the other instance's __eq__.
            return NotImplemented
        return self.id == other.id

    def __ne__(self, other):  # type: ignore[explicit-override] # FIXME
        return not self.__eq__(other=other)

    @override
    def __str__(self):
        return json.dumps(self.to_json_dict(), indent=2)

    @override
    def __hash__(self) -> int:  # type: ignore[override] # standard Python dicts are unhashable because of mutability
        """Overrides the default implementation"""
        _result_hash: int = hash(self.id)
        return _result_hash
